"""
SPI test for the CC3200 based boards.
"""

from machine import SPI
import os

mch = os.uname().machine
if "LaunchPad" in mch:
    spi_pins = ("GP14", "GP16", "GP30")
elif "WiPy" in mch:
    spi_pins = ("GP14", "GP16", "GP30")
else:
    raise Exception("Board not supported!")

spi = SPI(0, SPI.MASTER, baudrate=2000000, polarity=0, phase=0, firstbit=SPI.MSB, pins=spi_pins)
print(spi)
spi = SPI(baudrate=5000000)
print(spi)
spi = SPI(0, SPI.MASTER, baudrate=200000, bits=16, polarity=0, phase=0)
print(spi)
spi = SPI(0, SPI.MASTER, baudrate=10000000, polarity=0, phase=1)
print(spi)
spi = SPI(0, SPI.MASTER, baudrate=5000000, bits=32, polarity=1, phase=0)
print(spi)
spi = SPI(0, SPI.MASTER, baudrate=10000000, polarity=1, phase=1)
print(spi)
spi.init(baudrate=20000000, polarity=0, phase=0)
print(spi)
spi = SPI()
print(spi)
SPI(mode=SPI.MASTER)
SPI(mode=SPI.MASTER, pins=spi_pins)
SPI(id=0, mode=SPI.MASTER, polarity=0, phase=0, pins=("GP14", "GP16", "GP15"))
SPI(0, SPI.MASTER, polarity=0, phase=0, pins=("GP31", "GP16", "GP15"))

spi = SPI(0, SPI.MASTER, baudrate=10000000, polarity=0, phase=0, pins=spi_pins)
print(spi.write("123456") == 6)
buffer_r = bytearray(10)
print(spi.readinto(buffer_r) == 10)
print(spi.readinto(buffer_r, write=0x55) == 10)
read = spi.read(10)
print(len(read) == 10)
read = spi.read(10, write=0xFF)
print(len(read) == 10)
buffer_w = bytearray([1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
print(spi.write_readinto(buffer_w, buffer_r) == 10)
print(buffer_w == buffer_r)

# test all polaritiy and phase combinations
spi.init(polarity=1, phase=0, pins=None)
buffer_r = bytearray(10)
spi.write_readinto(buffer_w, buffer_r)
print(buffer_w == buffer_r)

spi.init(polarity=1, phase=1, pins=None)
buffer_r = bytearray(10)
spi.write_readinto(buffer_w, buffer_r)
print(buffer_w == buffer_r)

spi.init(polarity=0, phase=1, pins=None)
buffer_r = bytearray(10)
spi.write_readinto(buffer_w, buffer_r)
print(buffer_w == buffer_r)

# test 16 and 32 bit transfers
buffer_w = bytearray([1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2])
buffer_r = bytearray(12)
spi.init(SPI.MASTER, baudrate=10000000, bits=16, polarity=0, phase=0, pins=None)
print(spi.write_readinto(buffer_w, buffer_r) == 12)
print(buffer_w == buffer_r)

buffer_r = bytearray(12)
spi.init(SPI.MASTER, baudrate=10000000, bits=32, polarity=0, phase=0, pins=None)
print(spi.write_readinto(buffer_w, buffer_r) == 12)
print(buffer_w == buffer_r)

# check for memory leaks...
for i in range(0, 1000):
    spi = SPI(0, SPI.MASTER, baudrate=1000000)

# test deinit
spi = SPI(0, SPI.MASTER, baudrate=1000000)
spi.deinit()
print(spi)

spi = SPI(0, SPI.MASTER, baudrate=1000000)
# next ones must fail
try:
    spi = SPI(0, 10, baudrate=10000000, polarity=0, phase=0)
except:
    print("Exception")

try:
    spi = SPI(0, mode=SPI.MASTER, baudrate=10000000, polarity=1, phase=2)
except:
    print("Exception")

try:
    spi = SPI(1, mode=SPI.MASTER, baudrate=10000000, polarity=1, phase=1)
except:
    print("Exception")

try:
    spi = SPI(0, mode=SPI.MASTER, baudrate=2000000, polarity=2, phase=0)
except:
    print("Exception")

try:
    spi = SPI(0, mode=SPI.MASTER, baudrate=2000000, polarity=2, phase=0, firstbit=2)
except:
    print("Exception")

try:
    spi = SPI(0, mode=SPI.MASTER, baudrate=2000000, polarity=2, phase=0, pins=("GP1", "GP2"))
except:
    print("Exception")

try:
    spi = SPI(0, mode=SPI.MASTER, baudrate=2000000, polarity=0, phase=0, bits=9)
except:
    print("Exception")

spi.deinit()
try:
    spi.read(15)
except Exception:
    print("Exception")

try:
    spi.spi.readinto(buffer_r)
except Exception:
    print("Exception")

try:
    spi.spi.write("abc")
except Exception:
    print("Exception")

try:
    spi.write_readinto(buffer_w, buffer_r)
except Exception:
    print("Exception")

# reinitialization must work
spi.init(baudrate=500000)
print(spi)



import pyb

# test basic functionality
ext = pyb.ExtInt("X5", pyb.ExtInt.IRQ_RISING, pyb.Pin.PULL_DOWN, lambda l: print("line:", l))
ext.disable()
ext.enable()
print(ext.line())
ext.swint()

# test swint while disabled, then again after re-enabled
ext.disable()
ext.swint()
ext.enable()
ext.swint()

# disable now that the test is finished
ext.disable()



# test construction of array from array with float type

try:
    from array import array
except ImportError:
    print("SKIP")
    raise SystemExit

print(array("f", array("h", [1, 2])))
print(array("d", array("f", [1, 2])))



# This tests the performance of an instance class locals dict (importantly, that has all keys as qstrs)

# These are all shorter than 10 characters, so will be interned by the parser.
ZEN = [
    "the",
    "zen",
    "of",
    "python",
    "beautiful",
    "is",
    "better",
    "than",
    "ugly",
    "explicit",
    "is",
    "better",
    "than",
    "implicit",
    "simple",
    "is",
    "better",
    "than",
    "complex",
    "complex",
    "is",
    "better",
    "than",
    "complicate",
    "flat",
    "is",
    "better",
    "than",
    "nested",
    "sparse",
    "is",
    "better",
    "than",
    "dense",
    "readabilit",
    "counts",
    "special",
    "cases",
    "arent",
    "special",
    "enough",
    "to",
    "break",
    "the",
    "rules",
    "although",
    "practicali",
    "beats",
    "purity",
    "errors",
    "should",
    "never",
    "pass",
    "silently",
    "unless",
    "explicitly",
    "silenced",
    "in",
    "the",
    "face",
    "of",
    "ambiguity",
    "refuse",
    "the",
    "temptation",
    "to",
    "guess",
    "there",
    "should",
    "be",
    "one",
    "and",
    "preferably",
    "only",
    "one",
    "obvious",
    "way",
    "to",
    "do",
    "it",
    "although",
    "that",
    "way",
    "may",
    "not",
    "be",
    "obvious",
    "at",
    "first",
    "unless",
    "youre",
    "dutch",
    "now",
    "is",
    "better",
    "than",
    "never",
    "although",
    "never",
    "is",
    "often",
    "better",
    "than",
    "right",
    "now",
    "if",
    "the",
    "implementa",
    "is",
    "hard",
    "to",
    "explain",
    "its",
    "a",
    "bad",
    "idea",
    "if",
    "the",
    "implementa",
    "is",
    "easy",
    "to",
    "explain",
    "it",
    "may",
    "be",
    "a",
    "good",
    "idea",
    "namespaces",
    "are",
    "one",
    "honking",
    "great",
    "idea",
    "",
    "lets",
    "do",
    "more",
    "of",
    "those",
]


class A:
    pass


def test(niter):
    for _ in range(niter):
        a = A()
        for _ in range(niter):
            for word in ZEN:
                setattr(a, word, getattr(a, word, 0) + 1)

    return (
        getattr(a, "python"),
        getattr(a, "is"),
        getattr(a, "than"),
    )


###########################################################################
# Benchmark interface

bm_params = {
    (32, 10): (2,),
    (50, 10): (4,),
    (100, 10): (8,),
    (500, 10): (40,),
    (1000, 10): (80,),
    (5000, 10): (400,),
}


def bm_setup(params):
    (niter,) = params
    state = None

    def run():
        nonlocal state
        state = test(niter)

    def result():
        return niter, state

    return run, result



# This is hwconfig for "emulation" for cases when there's no real hardware.
# It just prints information to console.
class LEDClass:
    def __init__(self, id):
        self.id = "LED(%d):" % id

    def value(self, v):
        print(self.id, v)

    def on(self):
        self.value(1)

    def off(self):
        self.value(0)


LED = LEDClass(1)
LED2 = LEDClass(12)



# Test that integers format to exact values.
# This test requires at least 32-bit floats (won't work with 30-bit).

# Check that powers of 10 (that fit in float32) format correctly.
for i in range(31):
    # It works to 12 digits on all platforms *except* qemu, where
    # 10^11 comes out as 10000000820 or something.
    print(i, "{:.7g}".format(float("1e" + str(i))))



# test builtin min and max functions
try:
    min
    max
except:
    print("SKIP")
    raise SystemExit

print(min(0,1))
print(min(1,0))
print(min(0,-1))
print(min(-1,0))

print(max(0,1))
print(max(1,0))
print(max(0,-1))
print(max(-1,0))

print(min([1,2,4,0,-1,2]))
print(max([1,2,4,0,-1,2]))

# test with key function
lst = [2, 1, 3, 4]
print(min(lst, key=lambda x:x))
print(min(lst, key=lambda x:-x))
print(min(1, 2, 3, 4, key=lambda x:-x))
print(min(4, 3, 2, 1, key=lambda x:-x))
print(max(lst, key=lambda x:x))
print(max(lst, key=lambda x:-x))
print(max(1, 2, 3, 4, key=lambda x:-x))
print(max(4, 3, 2, 1, key=lambda x:-x))

# need at least 1 item in the iterable
try:
    min([])
except ValueError:
    print("ValueError")

# 'default' tests
print(min([1, 2, 3, 4, 5], default=-1))
print(min([], default=-1))
print(max([1, 2, 3, 4, 5], default=-1))
print(max([], default=-1))
# make sure it works with lazy iterables
# can't use Python generators here, as they're not supported
# byy native codegenerator.
print(min(enumerate([]), default=-10))



# Ping-pong GATT notifications between two devices.

from micropython import const
import time, machine, bluetooth

TIMEOUT_MS = 2000

_IRQ_CENTRAL_CONNECT = const(1)
_IRQ_CENTRAL_DISCONNECT = const(2)
_IRQ_PERIPHERAL_CONNECT = const(7)
_IRQ_PERIPHERAL_DISCONNECT = const(8)
_IRQ_GATTC_CHARACTERISTIC_RESULT = const(11)
_IRQ_GATTC_CHARACTERISTIC_DONE = const(12)
_IRQ_GATTC_NOTIFY = const(18)

# How long to run the test for.
_NUM_NOTIFICATIONS = const(50)

SERVICE_UUID = bluetooth.UUID("A5A5A5A5-FFFF-9999-1111-5A5A5A5A5A5A")
CHAR_UUID = bluetooth.UUID("00000000-1111-2222-3333-444444444444")
CHAR = (
    CHAR_UUID,
    bluetooth.FLAG_NOTIFY,
)
SERVICE = (
    SERVICE_UUID,
    (CHAR,),
)
SERVICES = (SERVICE,)

is_central = False

waiting_events = {}


def irq(event, data):
    if event == _IRQ_CENTRAL_CONNECT:
        waiting_events[event] = data[0]
    elif event == _IRQ_PERIPHERAL_CONNECT:
        waiting_events[event] = data[0]
    elif event == _IRQ_GATTC_CHARACTERISTIC_RESULT:
        # conn_handle, def_handle, value_handle, properties, uuid = data
        if data[-1] == CHAR_UUID:
            waiting_events[event] = data[2]
        else:
            return
    elif event == _IRQ_GATTC_NOTIFY:
        if is_central:
            conn_handle, value_handle, notify_data = data
            ble.gatts_notify(conn_handle, value_handle, b"central" + notify_data)

    if event not in waiting_events:
        waiting_events[event] = None


def wait_for_event(event, timeout_ms):
    t0 = time.ticks_ms()
    while time.ticks_diff(time.ticks_ms(), t0) < timeout_ms:
        if event in waiting_events:
            return waiting_events.pop(event)
        machine.idle()
    raise ValueError("Timeout waiting for {}".format(event))


# Acting in peripheral role.
def instance0():
    multitest.globals(BDADDR=ble.config("mac"))
    ((char_handle,),) = ble.gatts_register_services(SERVICES)
    print("gap_advertise")
    ble.gap_advertise(20_000, b"\x02\x01\x06\x04\xffMPY")
    multitest.next()
    try:
        # Wait for central to connect to us.
        conn_handle = wait_for_event(_IRQ_CENTRAL_CONNECT, TIMEOUT_MS)

        # Discover characteristics.
        ble.gattc_discover_characteristics(conn_handle, 1, 65535)
        value_handle = wait_for_event(_IRQ_GATTC_CHARACTERISTIC_RESULT, TIMEOUT_MS)
        wait_for_event(_IRQ_GATTC_CHARACTERISTIC_DONE, TIMEOUT_MS)

        # Give the central enough time to discover chars.
        time.sleep_ms(500)

        ticks_start = time.ticks_ms()

        for i in range(_NUM_NOTIFICATIONS):
            # Send a notification and wait for a response.
            ble.gatts_notify(conn_handle, value_handle, "peripheral" + str(i))
            wait_for_event(_IRQ_GATTC_NOTIFY, TIMEOUT_MS)

        ticks_end = time.ticks_ms()
        ticks_total = time.ticks_diff(ticks_end, ticks_start)
        multitest.output_metric(
            "Acknowledged {} notifications in {} ms. {} ms/notification.".format(
                _NUM_NOTIFICATIONS, ticks_total, ticks_total // _NUM_NOTIFICATIONS
            )
        )

        # Disconnect the central.
        print("gap_disconnect:", ble.gap_disconnect(conn_handle))
        wait_for_event(_IRQ_CENTRAL_DISCONNECT, TIMEOUT_MS)
    finally:
        ble.active(0)


# Acting in central role.
def instance1():
    global is_central
    is_central = True
    ((char_handle,),) = ble.gatts_register_services(SERVICES)
    multitest.next()
    try:
        # Connect to peripheral, with a short connection interval to reduce notify latency.
        print("gap_connect")
        ble.gap_connect(BDADDR[0], BDADDR[1], 2000, 12500, 12500)
        conn_handle = wait_for_event(_IRQ_PERIPHERAL_CONNECT, TIMEOUT_MS)

        # Discover characteristics.
        ble.gattc_discover_characteristics(conn_handle, 1, 65535)
        value_handle = wait_for_event(_IRQ_GATTC_CHARACTERISTIC_RESULT, TIMEOUT_MS)
        wait_for_event(_IRQ_GATTC_CHARACTERISTIC_DONE, TIMEOUT_MS)

        # The IRQ handler will respond to each notification.

        # Wait for the peripheral to disconnect us.
        wait_for_event(_IRQ_PERIPHERAL_DISCONNECT, 20000)
    finally:
        ble.active(0)


ble = bluetooth.BLE()
ble.active(1)
ble.irq(irq)



# MicroPython asyncio module, for use with webassembly port
# MIT license; Copyright (c) 2024 Damien P. George

from .core import *
from .funcs import wait_for, wait_for_ms, gather
from .event import Event
from .lock import Lock

__version__ = (3, 0, 0)



#!/usr/bin/env python3
"""
Build manifest generation for the NEAR Python contract compiler.
"""

import os
from pathlib import Path
from typing import Any, List, Optional, Set, Tuple

from .analyzer import MPY_STDLIB_PACKAGES, is_micropython_module
from .utils import console, find_site_packages


class ManifestGenerator:
    """Handles the generation of build manifests for NEAR Python contracts."""

    def __init__(
        self,
        contract_path: Path,
        imports: Set[str],
        exports: Set[str],
        venv_path: Path,
        build_dir: Path,
        single_file: bool = False,
    ):
        self.contract_path = contract_path
        self.imports = imports
        self.exports = exports
        self.venv_path = venv_path
        self.build_dir = build_dir
        self.contract_dir = contract_path.parent
        self.site_packages = self._get_site_packages()
        self.excluded_stdlib_packages = self._get_excluded_stdlib_packages()
        self.gitignore_spec = self._load_gitignore_spec()
        self.single_file = single_file

    def _get_site_packages(self) -> Path:
        """Get the site-packages directory from the virtual environment."""
        site_packages = find_site_packages(self.venv_path)
        if not site_packages:
            console.print(
                f"[red]Error: Could not find site-packages in {self.venv_path}"
            )
            import sys

            sys.exit(1)
        return site_packages

    def _get_excluded_stdlib_packages(self) -> List[str]:
        """Get excluded stdlib packages from pyproject.toml."""
        pyproject_path = self.contract_dir / "pyproject.toml"
        excluded_packages = []

        if pyproject_path.is_file():
            try:
                import tomllib

                with open(pyproject_path, "rb") as file:
                    pyproject_data = tomllib.load(file)
                excluded_packages = (
                    pyproject_data.get("tool", {})
                    .get("near-py-tool", {})
                    .get("exclude-micropython-stdlib-packages", [])
                )
            except Exception as e:
                console.print(
                    f"[yellow]Warning: Could not read exclusions from pyproject.toml: {e}"
                )

        if excluded_packages:
            console.print(
                f"Excluding MicroPython stdlib packages: {', '.join(excluded_packages)}"
            )

        return excluded_packages

    def _load_gitignore_spec(self) -> Optional[Any]:
        """Load gitignore patterns if available."""
        try:
            import pathspec
        except ImportError:
            console.print(
                "[yellow]pathspec library not found, gitignore filtering disabled[/]"
            )
            console.print("[yellow]Install with: pip install pathspec[/]")
            return None

        gitignore_path = self.contract_dir / ".gitignore"
        if not gitignore_path.exists():
            return None

        with open(gitignore_path, "r") as gitignore_file:
            gitignore_patterns = gitignore_file.read().splitlines()

            # Add common Python patterns if not already specified
            default_patterns = [
                "__pycache__/",
                "*.py[cod]",
                "*$py.class",
                "*.so",
                "build/",
                "dist/",
            ]
            for pattern in default_patterns:
                if pattern not in gitignore_patterns:
                    gitignore_patterns.append(pattern)

            spec = pathspec.PathSpec.from_lines(
                pathspec.patterns.GitWildMatchPattern, gitignore_patterns
            )
            console.print("[cyan]Using .gitignore patterns for module filtering[/]")
            return spec

    def find_local_modules(self) -> List[Path]:
        """Find all local Python modules in the contract directory, respecting .gitignore."""
        local_modules: List[Path] = []

        # If single_file mode is enabled, skip looking for local modules
        if self.single_file:
            console.print("[cyan]Single file mode: skipping local module discovery[/]")
            return local_modules

        always_exclude = [".git", ".venv", "venv", "__pycache__", "build"]

        console.print("[cyan]Scanning for local Python modules...[/]")

        for py_file in self.contract_dir.glob("**/*.py"):
            # Skip the main contract file and generated files
            if py_file.name == self.contract_path.name or py_file.name.endswith(
                ("_with_metadata.py", "_with_abi.py")
            ):
                continue

            # Get relative path for gitignore matching
            rel_path = py_file.relative_to(self.contract_dir)
            rel_path_str = str(rel_path).replace("\\", "/")

            # Skip based on always_exclude patterns
            if any(exclude in str(py_file) for exclude in always_exclude):
                continue

            # Skip if matches gitignore patterns
            if (
                self.gitignore_spec
                and hasattr(self.gitignore_spec, "match_file")
                and self.gitignore_spec.match_file(rel_path_str)
            ):
                console.print(
                    f"  [dim yellow]Ignoring (gitignore match): {rel_path}[/]"
                )
                continue

            local_modules.append(rel_path)
            console.print(f"  [dim]Found local module: {rel_path}[/]")

        if not local_modules:
            console.print(
                "[yellow]No local modules found besides the main contract file[/]"
            )
        else:
            console.print(
                f"[green]Found {len(local_modules)} local modules to include[/]"
            )

        return local_modules

    def process_external_dependencies(
        self, local_modules: List[Path]
    ) -> List[Tuple[str, str]]:
        """Process external dependencies from imports list."""
        external_modules = {
            name.split(".")[0]
            for name in self.imports
            if not is_micropython_module(name)
        }

        external_deps = []
        missing_modules = []

        for base_module in sorted(external_modules):
            # Check if it's a local module
            local_module_file = self.contract_dir / f"{base_module}.py"
            local_module_dir = self.contract_dir / base_module
            local_module_init = local_module_dir / "__init__.py"

            if local_module_file.exists() or (
                local_module_dir.is_dir() and local_module_init.exists()
            ):
                continue  # Local module, already handled

            # Check in site-packages
            module_dir = self.site_packages / base_module
            module_file = self.site_packages / f"{base_module}.py"

            if module_dir.is_dir():
                external_deps.append((base_module, "package"))
            elif module_file.exists():
                external_deps.append((base_module, "module"))
            else:
                missing_modules.append(base_module)
                console.print(
                    f"[yellow]Warning: Could not find module {base_module} in {self.site_packages}"
                )

        return external_deps

    def write_manifest(
        self, local_modules: List[Path], external_deps: List[Tuple[str, str]]
    ) -> Path:
        """Write the manifest file."""
        manifest_path = self.build_dir / "manifest.py"

        with open(manifest_path, "w") as f:
            f.write("# THIS FILE IS GENERATED, DO NOT EDIT\n\n")

            # Add stdlib packages
            included_stdlib_packages = set(MPY_STDLIB_PACKAGES) - set(
                self.excluded_stdlib_packages
            )
            f.write(
                "\n".join(
                    f'require("{module}")'
                    for module in sorted(included_stdlib_packages)
                )
            )

            # Add typing modules
            f.write("\n\n# Typing modules\n")
            f.write(
                "\n".join(
                    f'module("{mod}.py", base_path="$(PORT_DIR)/extra/typing")'
                    for mod in ["typing", "typing_extensions"]
                )
            )

            # Add local modules
            if local_modules:
                f.write("\n\n# Local modules\n")
                contract_rel_path = os.path.relpath(
                    self.contract_dir, manifest_path.parent
                ).replace("\\", "/")

                for rel_path in sorted(local_modules):
                    f.write(f'module("{rel_path}", base_path="{contract_rel_path}")\n')

            # Add external dependencies
            if external_deps:
                f.write("\n\n# External dependencies\n")
                rel_path_str = os.path.relpath(
                    self.site_packages, manifest_path.parent
                ).replace("\\", "/")

                for module_name, module_type in external_deps:
                    if module_type == "package":
                        f.write(
                            f'package("{module_name}", base_path="{rel_path_str}")\n'
                        )
                    else:
                        f.write(
                            f'module("{module_name}.py", base_path="{rel_path_str}")\n'
                        )

            # Add contract file
            f.write("\n\n# Contract\n")
            f.write(f'module("{self.contract_path.name}", base_path="..")')

        return manifest_path

    def write_wrappers(self) -> Path:
        """Generate export wrappers file."""
        wrappers_path = self.build_dir / "export_wrappers.c"

        with open(wrappers_path, "w") as f:
            f.write("/* Generated export wrappers for NEAR contract */\n\n")
            f.write(
                "void run_frozen_fn(const char *file_name, const char *fn_name);\n\n"
            )

            for export in sorted(self.exports):
                f.write(f"void {export}() {{\n")
                f.write(
                    f'    run_frozen_fn("{self.contract_path.name}", "{export}");\n'
                )
                f.write("}\n\n")

        return wrappers_path

    def generate(self) -> Tuple[Path, Path]:
        """Generate all build files."""
        local_modules = self.find_local_modules()
        external_deps = self.process_external_dependencies(local_modules)

        console.print("[cyan]Generating build files...[/]", end="")
        manifest_path = self.write_manifest(local_modules, external_deps)
        wrappers_path = self.write_wrappers()
        console.print(" done")

        return manifest_path, wrappers_path


def prepare_build_files(
    contract_path: Path,
    imports: Set[str],
    exports: Set[str],
    venv_path: Path,
    build_dir: Path,
    single_file: bool = False,
) -> Tuple[Path, Path]:
    """
    Generate manifest and wrapper files for NEAR contract compilation.

    Args:
        contract_path: Path to the contract file
        imports: Set of imported module names
        exports: Set of exported function names
        venv_path: Path to the virtual environment
        build_dir: Path to the build directory
        single_file: Whether to skip local module discovery and compile only the specified file

    Returns:
        Tuple of (manifest_path, wrappers_path)
    """
    generator = ManifestGenerator(
        contract_path, imports, exports, venv_path, build_dir, single_file
    )
    return generator.generate()



# MIT license; Copyright (c) 2021 Jim Mussared

# This is a WIP client for l2cap_file_server.py. See that file for more
# information.

import sys

# ruff: noqa: E402
sys.path.append("")

from micropython import const

import asyncio
import aioble
import bluetooth

import random
import struct

_FILE_SERVICE_UUID = bluetooth.UUID(0x1234)
_CONTROL_CHARACTERISTIC_UUID = bluetooth.UUID(0x1235)


_COMMAND_SEND = const(0)
_COMMAND_RECV = const(1)  # Not yet implemented.
_COMMAND_LIST = const(2)
_COMMAND_SIZE = const(3)
_COMMAND_DONE = const(4)

_STATUS_OK = const(0)
_STATUS_NOT_IMPLEMENTED = const(1)
_STATUS_NOT_FOUND = const(2)

_L2CAP_PSN = const(22)
_L2CAP_MTU = const(128)


class FileClient:
    def __init__(self, device):
        self._device = device
        self._connection = None
        self._seq = 1

    async def connect(self):
        try:
            print("Connecting to", self._device)
            self._connection = await self._device.connect()
        except asyncio.TimeoutError:
            print("Timeout during connection")
            return

        try:
            print("Discovering...")
            file_service = await self._connection.service(_FILE_SERVICE_UUID)
            self._control_characteristic = await file_service.characteristic(
                _CONTROL_CHARACTERISTIC_UUID
            )
        except asyncio.TimeoutError:
            print("Timeout discovering services/characteristics")
            return

        print("Connecting channel")
        self._channel = await self._connection.l2cap_connect(_L2CAP_PSN, _L2CAP_MTU)

    async def _command(self, cmd, data):
        send_seq = self._seq
        await self._control_characteristic.write(struct.pack("<BB", cmd, send_seq) + data)
        self._seq += 1
        return send_seq

    async def size(self, path):
        print("Getting size")
        send_seq = await self._command(_COMMAND_SIZE, path.encode())

        data = await self._control_characteristic.notified()
        if len(data) != 6:
            raise RuntimeError("Invalid response")

        seq, status, size = struct.unpack("<BBI", data)
        if seq != send_seq:
            raise RuntimeError("Wrong reply")

        print("result:", seq, status, size)
        return size

    async def download(self, path, dest):
        size = await self.size(path)

        await self._command(_COMMAND_SEND, path.encode())

        with open(dest, "wb") as f:  # noqa: ASYNC101
            total = 0
            buf = bytearray(self._channel.our_mtu)
            mv = memoryview(buf)
            while total < size:
                n = await self._channel.recvinto(buf)
                f.write(mv[:n])
                total += n

    async def list(self, path):
        await self._command(_COMMAND_LIST, path.encode())
        results = bytearray()
        buf = bytearray(self._channel.our_mtu)
        mv = memoryview(buf)
        while True:
            n = await self._channel.recvinto(buf)
            results += mv[:n]
            if results[len(results) - 1] == ord("\n") and results[len(results) - 2] == ord("\n"):
                break
        print(results.decode().split("\n"))

    async def disconnect(self):
        if self._connection:
            await self._connection.disconnect()


async def main():
    async with aioble.scan(5000, 30000, 30000, active=True) as scanner:
        async for result in scanner:
            if result.name() == "mpy-file" and _FILE_SERVICE_UUID in result.services():
                device = result.device
                break
        else:
            print("File server not found")
            return

    client = FileClient(device)

    await client.connect()
    print(await client.size("demo/file.txt"))
    print(await client.size("demo/notfound.bin"))

    await client.download("demo/file.txt", "download.txt")

    await client.list("demo")

    await client.disconnect()


asyncio.run(main())



# MIT license; Copyright (c) 2023-2024 Angus Gratton
from micropython import const
import time
import usb.device
from usb.device.hid import HIDInterface

_INTERFACE_PROTOCOL_KEYBOARD = const(0x01)

_KEY_ARRAY_LEN = const(6)  # Size of HID key array, must match report descriptor
_KEY_REPORT_LEN = const(_KEY_ARRAY_LEN + 2)  # Modifier Byte + Reserved Byte + Array entries


class KeyboardInterface(HIDInterface):
    # Synchronous USB keyboard HID interface

    def __init__(self):
        super().__init__(
            _KEYBOARD_REPORT_DESC,
            set_report_buf=bytearray(1),
            protocol=_INTERFACE_PROTOCOL_KEYBOARD,
            interface_str="MicroPython Keyboard",
        )
        self._key_reports = [
            bytearray(_KEY_REPORT_LEN),
            bytearray(_KEY_REPORT_LEN),
        ]  # Ping/pong report buffers
        self.numlock = False

    def on_set_report(self, report_data, _report_id, _report_type):
        self.on_led_update(report_data[0])

    def on_led_update(self, led_mask):
        # Override to handle keyboard LED updates. led_mask is bitwise ORed
        # together values as defined in LEDCode.
        pass

    def send_keys(self, down_keys, timeout_ms=100):
        # Update the state of the keyboard by sending a report with down_keys
        # set, where down_keys is an iterable (list or similar) of integer
        # values such as the values defined in KeyCode.
        #
        # Will block for up to timeout_ms if a previous report is still
        # pending to be sent to the host. Returns True on success.

        r, s = self._key_reports  # next report buffer to send, spare report buffer
        r[0] = 0  # modifier byte
        i = 2  # index for next key array item to write to
        for k in down_keys:
            if k < 0:  # Modifier key
                r[0] |= -k
            elif i < _KEY_REPORT_LEN:
                r[i] = k
                i += 1
            else:  # Excess rollover! Can't report
                r[0] = 0
                for i in range(2, _KEY_REPORT_LEN):
                    r[i] = 0xFF
                break

        while i < _KEY_REPORT_LEN:
            r[i] = 0
            i += 1

        if self.send_report(r, timeout_ms):
            # Swap buffers if the previous one is newly queued to send, so
            # any subsequent call can't modify that buffer mid-send
            self._key_reports[0] = s
            self._key_reports[1] = r
            return True
        return False


# HID keyboard report descriptor
#
# From p69 of http://www.usb.org/developers/devclass_docs/HID1_11.pdf
#
# fmt: off
_KEYBOARD_REPORT_DESC = (
    b'\x05\x01'     # Usage Page (Generic Desktop),
        b'\x09\x06'     # Usage (Keyboard),
    b'\xA1\x01'     # Collection (Application),
        b'\x05\x07'         # Usage Page (Key Codes);
            b'\x19\xE0'         # Usage Minimum (224),
            b'\x29\xE7'         # Usage Maximum (231),
            b'\x15\x00'         # Logical Minimum (0),
            b'\x25\x01'         # Logical Maximum (1),
            b'\x75\x01'         # Report Size (1),
            b'\x95\x08'         # Report Count (8),
            b'\x81\x02'         # Input (Data, Variable, Absolute), ;Modifier byte
            b'\x95\x01'         # Report Count (1),
            b'\x75\x08'         # Report Size (8),
            b'\x81\x01'         # Input (Constant), ;Reserved byte
            b'\x95\x05'         # Report Count (5),
            b'\x75\x01'         # Report Size (1),
        b'\x05\x08'         # Usage Page (Page# for LEDs),
            b'\x19\x01'         # Usage Minimum (1),
            b'\x29\x05'         # Usage Maximum (5),
            b'\x91\x02'         # Output (Data, Variable, Absolute), ;LED report
            b'\x95\x01'         # Report Count (1),
            b'\x75\x03'         # Report Size (3),
            b'\x91\x01'         # Output (Constant), ;LED report padding
            b'\x95\x06'         # Report Count (6),
            b'\x75\x08'         # Report Size (8),
            b'\x15\x00'         # Logical Minimum (0),
            b'\x25\x65'         # Logical Maximum(101),
        b'\x05\x07'         # Usage Page (Key Codes),
            b'\x19\x00'         # Usage Minimum (0),
            b'\x29\x65'         # Usage Maximum (101),
            b'\x81\x00'         # Input (Data, Array), ;Key arrays (6 bytes)
    b'\xC0'     # End Collection
)
# fmt: on


# Standard HID keycodes, as a pseudo-enum class for easy access
#
# Modifier keys are encoded as negative values
class KeyCode:
    A = 4
    B = 5
    C = 6
    D = 7
    E = 8
    F = 9
    G = 10
    H = 11
    I = 12
    J = 13
    K = 14
    L = 15
    M = 16
    N = 17
    O = 18
    P = 19
    Q = 20
    R = 21
    S = 22
    T = 23
    U = 24
    V = 25
    W = 26
    X = 27
    Y = 28
    Z = 29
    N1 = 30  # Standard number row keys
    N2 = 31
    N3 = 32
    N4 = 33
    N5 = 34
    N6 = 35
    N7 = 36
    N8 = 37
    N9 = 38
    N0 = 39
    ENTER = 40
    ESCAPE = 41
    BACKSPACE = 42
    TAB = 43
    SPACE = 44
    MINUS = 45  # - _
    EQUAL = 46  # = +
    OPEN_BRACKET = 47  # [ {
    CLOSE_BRACKET = 48  # ] }
    BACKSLASH = 49  # \ |
    HASH = 50  # # ~
    SEMICOLON = 51  # ; :
    QUOTE = 52  # ' "
    GRAVE = 53  # ` ~
    COMMA = 54  # , <
    DOT = 55  # . >
    SLASH = 56  # / ?
    CAPS_LOCK = 57
    F1 = 58
    F2 = 59
    F3 = 60
    F4 = 61
    F5 = 62
    F6 = 63
    F7 = 64
    F8 = 65
    F9 = 66
    F10 = 67
    F11 = 68
    F12 = 69
    PRINTSCREEN = 70
    SCROLL_LOCK = 71
    PAUSE = 72
    INSERT = 73
    HOME = 74
    PAGEUP = 75
    DELETE = 76
    END = 77
    PAGEDOWN = 78
    RIGHT = 79  # Arrow keys
    LEFT = 80
    DOWN = 81
    UP = 82
    KP_NUM_LOCK = 83
    KP_DIVIDE = 84
    KP_AT = 85
    KP_MULTIPLY = 85
    KP_MINUS = 86
    KP_PLUS = 87
    KP_ENTER = 88
    KP_1 = 89
    KP_2 = 90
    KP_3 = 91
    KP_4 = 92
    KP_5 = 93
    KP_6 = 94
    KP_7 = 95
    KP_8 = 96
    KP_9 = 97
    KP_0 = 98

    # HID modifier values (negated to allow them to be passed along with the normal keys)
    LEFT_CTRL = -0x01
    LEFT_SHIFT = -0x02
    LEFT_ALT = -0x04
    LEFT_UI = -0x08
    RIGHT_CTRL = -0x10
    RIGHT_SHIFT = -0x20
    RIGHT_ALT = -0x40
    RIGHT_UI = -0x80


# HID LED values
class LEDCode:
    NUM_LOCK = 0x01
    CAPS_LOCK = 0x02
    SCROLL_LOCK = 0x04
    COMPOSE = 0x08
    KANA = 0x10



# test time.time_ns()

try:
    import time

    time.sleep_us
    time.time_ns
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


t0 = time.time_ns()
time.sleep_us(5000)
t1 = time.time_ns()

# Check that time_ns increases.
print(t0 < t1)

# Check that time_ns counts correctly, but be very lenient with the bounds (2ms to 50ms).
if 2000000 < t1 - t0 < 50000000:
    print(True)
else:
    print(t0, t1, t1 - t0)

# Check that time.time() and time.time_ns() are within a second of each other.
# Note that time.time() may return an int or float.
for _ in range(10):
    t_s, t_ns = time.time(), time.time_ns()
    print(abs(t_s * 1_000 - t_ns // 1_000_000) <= 1_000)
    time.sleep_us(100_000)



"""
I2C test for the CC3200 based boards.
A MPU-9150 sensor must be connected to the I2C bus.
"""

from machine import I2C
import os
import time

mch = os.uname().machine
if "LaunchPad" in mch:
    i2c_pins = ("GP11", "GP10")
elif "WiPy" in mch:
    i2c_pins = ("GP15", "GP10")
else:
    raise Exception("Board not supported!")

i2c = I2C(0, I2C.MASTER, baudrate=400000)
# try initing without the peripheral id
i2c = I2C()
print(i2c)
i2c = I2C(mode=I2C.MASTER, baudrate=50000, pins=i2c_pins)
print(i2c)

i2c = I2C(0, I2C.MASTER, baudrate=100000)
print(i2c)
i2c = I2C(0, mode=I2C.MASTER, baudrate=400000)
print(i2c)
i2c = I2C(0, mode=I2C.MASTER, baudrate=400000, pins=i2c_pins)
print(i2c)

addr = i2c.scan()[0]
print(addr)

reg = bytearray(1)
reg2 = bytearray(2)
reg2_r = bytearray(2)

# reset the sensor
reg[0] |= 0x80
print(1 == i2c.writeto_mem(addr, 107, reg))
time.sleep_ms(100)  # wait for the sensor to reset...

print(1 == i2c.readfrom_mem_into(addr, 107, reg))  # read the power management register 1
print(0x40 == reg[0])

# now just read one byte
data = i2c.readfrom_mem(addr, 117, 1)  # read the "who am I?" register
print(0x68 == data[0])
print(len(data) == 1)
print(1 == i2c.readfrom_mem_into(addr, 117, reg))  # read the "who am I?" register again
print(0x68 == reg[0])

# now try reading two bytes
data = i2c.readfrom_mem(addr, 116, 2)  # read the "who am I?" register
print(0x68 == data[1])
print(data == b"\x00\x68")
print(len(data) == 2)
print(2 == i2c.readfrom_mem_into(addr, 116, reg2))  # read the "who am I?" register again
print(0x68 == reg2[1])
print(reg2 == b"\x00\x68")

print(1 == i2c.readfrom_mem_into(addr, 107, reg))  # read the power management register 1
print(0x40 == reg[0])
# clear the sleep bit
reg[0] = 0
print(1 == i2c.writeto_mem(addr, 107, reg))
# read it back
i2c.readfrom_mem_into(addr, 107, reg)
print(0 == reg[0])

# set the sleep bit
reg[0] = 0x40
print(1 == i2c.writeto_mem(addr, 107, reg))
# read it back
i2c.readfrom_mem_into(addr, 107, reg)
print(0x40 == reg[0])

# reset the sensor
reg[0] |= 0x80
print(1 == i2c.writeto_mem(addr, 107, reg))
time.sleep_ms(100)  # wait for the sensor to reset...

# now read and write two register at a time
print(2 == i2c.readfrom_mem_into(addr, 107, reg2))
print(0x40 == reg2[0])
print(0x00 == reg2[1])
# clear the sleep bit
reg2[0] = 0
# set some other bits
reg2[1] |= 0x03
print(2 == i2c.writeto_mem(addr, 107, reg2))
# read it back
i2c.readfrom_mem_into(addr, 107, reg2_r)
print(reg2 == reg2_r)

# reset the sensor
reg[0] = 0x80
print(1 == i2c.writeto_mem(addr, 107, reg))
time.sleep_ms(100)  # wait for the sensor to reset...

# try some raw read and writes
reg[0] = 117  # register address
print(1 == i2c.writeto(addr, reg, stop=False))  # just write the register address
# now read
print(1 == i2c.readfrom_into(addr, reg))
print(reg[0] == 0x68)
reg[0] = 117  # register address
print(1 == i2c.writeto(addr, reg, stop=False))  # just write the register address
# now read
print(0x68 == i2c.readfrom(addr, 1)[0])

i2c.readfrom_mem_into(addr, 107, reg2)
print(0x40 == reg2[0])
print(0x00 == reg2[1])

reg2[0] = 107  # register address
reg2[1] = 0
print(2 == i2c.writeto(addr, reg2, stop=True))  # write the register address and the data
i2c.readfrom_mem_into(addr, 107, reg)  # check it back
print(reg[0] == 0)

# check for memory leaks...
for i in range(0, 1000):
    i2c = I2C(0, I2C.MASTER, baudrate=100000)

# test deinit
i2c = I2C(0, I2C.MASTER, baudrate=100000)
i2c.deinit()
print(i2c)

# next ones should raise
try:
    i2c.scan()
except Exception:
    print("Exception")

try:
    i2c.readfrom(addr, 1)
except Exception:
    print("Exception")

try:
    i2c.readfrom_into(addr, reg)
except Exception:
    print("Exception")

try:
    i2c.readfrom_mem_into(addr, 107, reg)
except Exception:
    print("Exception")

try:
    i2c.writeto(addr, reg, stop=False)
except Exception:
    print("Exception")

try:
    i2c.writeto_mem(addr, 107, reg)
except Exception:
    print("Exception")

try:
    i2c.readfrom_mem(addr, 116, 2)
except Exception:
    print("Exception")

try:
    I2C(1, I2C.MASTER, baudrate=100000)
except Exception:
    print("Exception")

# reinitialization must work
i2c.init(baudrate=400000)
print(i2c)



print("".replace("a", "b"))
print("aaa".replace("b", "c"))
print("aaa".replace("a", "b", 0))
print("aaa".replace("a", "b", -5))
print("asdfasdf".replace("a", "b"))
print("aabbaabbaabbaa".replace("aa", "cc", 3))
print("a".replace("aa", "bb"))
print("testingtesting".replace("ing", ""))
print("testINGtesting".replace("ing", "ING!"))

print("".replace("", "1"))
print("A".replace("", "1"))
print("AB".replace("", "1"))
print("AB".replace("", "12"))

try:
    'abc'.replace(1, 2)
except TypeError:
    print('TypeError')

try:
    'abc'.replace('1', 2)
except TypeError:
    print('TypeError')



metadata(
    version="0.2.0",
    description="Common networking packages for all network-capable deployments of MicroPython.",
)

require("mip")
require("ntptime")
require("ssl")
require("requests")
require("webrepl")

# Provide urequests (which just forwards to requests) for backwards
# compatibility.
require("urequests")



# add a bytearray to itself
# This is not supported by CPython as of 3.11.18.

b = bytearray(b"123456789")
for _ in range(4):
    c = bytearray(b)  # extra allocation increases chance 'b' has to relocate
    b += b
print(b)



# Threads should be semi-cooperative, to the point where one busy
# thread can't starve out another.
#
# (Note on ports without the GIL this one should always be true, on ports with GIL it's
# a test of the GIL behaviour.)

import _thread
import sys
from time import ticks_ms, ticks_diff, sleep_ms


done = False

ITERATIONS = 5
SLEEP_MS = 250
MAX_DELTA = 30

if sys.platform in ("win32", "linux", "darwin"):
    # Conventional operating systems get looser timing restrictions
    SLEEP_MS = 300
    MAX_DELTA = 100


def busy_thread():
    while not done:
        pass


def test_sleeps():
    global done
    ok = True
    for _ in range(ITERATIONS):
        t0 = ticks_ms()
        sleep_ms(SLEEP_MS)
        t1 = ticks_ms()
        d = ticks_diff(t1, t0)
        if d < SLEEP_MS - MAX_DELTA or d > SLEEP_MS + MAX_DELTA:
            print("Slept too long ", d)
            ok = False
    print("OK" if ok else "Not OK")
    done = True


# make the thread the busy one, and check sleep time on main task
_thread.start_new_thread(busy_thread, ())
test_sleeps()

sleep_ms(100)
done = False

# now swap them
_thread.start_new_thread(test_sleeps, ())
busy_thread()



# MicroPython asyncio module
# MIT license; Copyright (c) 2019-2020 Damien P. George

from . import core


# Event class for primitive events that can be waited on, set, and cleared
class Event:
    def __init__(self):
        self.state = False  # False=unset; True=set
        self.waiting = core.TaskQueue()  # Queue of Tasks waiting on completion of this event

    def is_set(self):
        return self.state

    def set(self):
        # Event becomes set, schedule any tasks waiting on it
        # Note: This must not be called from anything except the thread running
        # the asyncio loop (i.e. neither hard or soft IRQ, or a different thread).
        while self.waiting.peek():
            core._task_queue.push(self.waiting.pop())
        self.state = True

    def clear(self):
        self.state = False

    # async
    def wait(self):
        if not self.state:
            # Event not set, put the calling task on the event's waiting queue
            self.waiting.push(core.cur_task)
            # Set calling task's data to the event's queue so it can be removed if needed
            core.cur_task.data = self.waiting
            yield
        return True


# MicroPython-extension: This can be set from outside the asyncio event loop,
# such as other threads, IRQs or scheduler context. Implementation is a stream
# that asyncio will poll until a flag is set.
# Note: Unlike Event, this is self-clearing after a wait().
try:
    import io

    class ThreadSafeFlag(io.IOBase):
        def __init__(self):
            self.state = 0

        def ioctl(self, req, flags):
            if req == 3:  # MP_STREAM_POLL
                return self.state * flags
            return -1  # Other requests are unsupported

        def set(self):
            self.state = 1

        def clear(self):
            self.state = 0

        async def wait(self):
            if not self.state:
                yield core._io_queue.queue_read(self)
            self.state = 0

except ImportError:
    pass



import base64

b = base64.b64encode(b"zlutoucky kun upel dabelske ody")
print(b)

if b != b"emx1dG91Y2t5IGt1biB1cGVsIGRhYmVsc2tlIG9keQ==":
    raise Exception("Error")

d = base64.b64decode(b)
print(d)

if d != b"zlutoucky kun upel dabelske ody":
    raise Exception("Error")

base64.test()

binary = b"\x99\x10\xaa"
b = base64.b64encode(binary)
if b != b"mRCq":
    raise Exception("Error")

d = base64.b64decode(b)
print(d)
if d != binary:
    raise Exception("Error")

d = base64.b32encode(b"zlutoucky kun upel dabelske ody")
if d != b"PJWHK5DPOVRWW6JANN2W4IDVOBSWYIDEMFRGK3DTNNSSA33EPE======":
    raise Exception("Error")

print(d)
b = base64.b32decode(d)
if b != b"zlutoucky kun upel dabelske ody":
    raise Exception("Error")

print("OK")



# test the __globals__ attribute of a function


def foo():
    pass


if not hasattr(foo, "__globals__"):
    print("SKIP")
    raise SystemExit

print(type(foo.__globals__))
print(foo.__globals__ is globals())

foo.__globals__["bar"] = 123
print(bar)

try:
    foo.__globals__ = None
except AttributeError:
    print("AttributeError")

# test closures have the __globals__ attribute


def outer():
    x = 1

    def inner():
        return x

    return inner


print(outer.__globals__ is globals())
print(outer().__globals__ is globals())



# This module just allows `import uasyncio` to work. It lazy-loads from
# `asyncio` without duplicating its globals dict.


def __getattr__(attr):
    import asyncio

    return getattr(asyncio, attr)



# Example of a HTTPS client working with non-blocking sockets.
#
# Non-blocking SSL streams works differently in MicroPython compared to CPython.  In
# CPython a write to an SSLSocket may raise ssl.SSLWantReadError.  In MicroPython an
# SSLSocket behaves like a normal socket/stream and can be polled for reading/writing.

from errno import EINPROGRESS
import select
import socket
import ssl


def connect_nonblocking(sock, addr):
    sock.setblocking(False)
    try:
        sock.connect(addr)
    except OSError as er:
        if er.errno != EINPROGRESS:
            raise er


def write_nonblocking(poller, sock, data):
    poller.register(sock, select.POLLOUT)
    while data:
        poller.poll()
        n = sock.write(data)
        print("Wrote:", n)
        if n is not None:
            data = data[n:]


def read_nonblocking(poller, sock, n):
    poller.register(sock, select.POLLIN)
    poller.poll()
    data = sock.read(n)
    print("Read:", len(data))
    return data


def main(url, addr_family=0):
    # Split the given URL into components.
    proto, _, host, path = url.split(b"/", 3)
    assert proto == b"https:"

    # Note: this getaddrinfo() call is blocking!
    ai = socket.getaddrinfo(host, 443, addr_family, socket.SOCK_STREAM)[0]
    addr = ai[-1]
    print("Connect address:", addr)

    # Create a TCP socket and connect to the server in non-blocking mode.
    sock = socket.socket(ai[0], ai[1], ai[2])
    connect_nonblocking(sock, addr)

    # Wrap the TCP socket in an SSL stream in non-blocking mode.
    ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    sock = ctx.wrap_socket(sock, server_hostname=host, do_handshake_on_connect=False)
    sock.setblocking(False)

    # Create an object to poll the SSL stream for readability/writability.
    poller = select.poll()

    # Send the HTTP request on the SSL stream.
    request = b"GET /%s HTTP/1.0\r\nHost: %s\r\n\r\n" % (path, host)
    write_nonblocking(poller, sock, request)

    # Receive the HTTP response from the SSL stream.
    response = read_nonblocking(poller, sock, 1000)
    for line in response.split(b"\n"):
        print(line)

    # Close the SSL stream.  This will also close the underlying TCP socket.
    sock.close()


main(b"https://micropython.org/ks/test.html")



# Test uctypes array, load and store, with array size > 1

try:
    import uctypes
except ImportError:
    print("SKIP")
    raise SystemExit

N = 5

for endian in ("NATIVE", "LITTLE_ENDIAN", "BIG_ENDIAN"):
    for type_ in ("INT8", "UINT8", "INT16", "UINT16", "INT32", "UINT32", "INT64", "UINT64"):
        desc = {"arr": (uctypes.ARRAY | 0, getattr(uctypes, type_) | N)}
        sz = uctypes.sizeof(desc)
        data = bytearray(sz)
        s = uctypes.struct(uctypes.addressof(data), desc, getattr(uctypes, endian))
        for i in range(N):
            s.arr[i] = i - 2
        print(endian, type_, sz, *(s.arr[i] for i in range(N)))



# test VFS functionality with a user-defined filesystem
# also tests parts of io.IOBase implementation

import sys

try:
    import io, vfs

    io.IOBase
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class UserFile(io.IOBase):
    buffer_size = 16

    def __init__(self, mode, data):
        assert isinstance(data, bytes)
        self.is_text = mode.find("b") == -1
        self.data = data
        self.pos = 0

    def read(self):
        if self.is_text:
            return str(self.data, "utf8")
        else:
            return self.data

    def readinto(self, buf):
        assert not self.is_text
        n = 0
        while n < len(buf) and self.pos < len(self.data):
            buf[n] = self.data[self.pos]
            n += 1
            self.pos += 1
        return n

    def ioctl(self, req, arg):
        print("ioctl", req, arg)
        if req == 4:  # MP_STREAM_CLOSE
            return 0
        if req == 11:  # MP_STREAM_GET_BUFFER_SIZE
            return UserFile.buffer_size
        return -1


class UserFS:
    def __init__(self, files):
        self.files = files

    def mount(self, readonly, mksfs):
        pass

    def umount(self):
        pass

    def stat(self, path):
        print("stat", path)
        if path in self.files:
            return (32768, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        raise OSError

    def open(self, path, mode):
        print("open", path, mode)
        return UserFile(mode, self.files[path])


# create and mount a user filesystem
user_files = {
    "/data.txt": b"some data in a text file",
    "/usermod1.py": b"print('in usermod1')\nimport usermod2",
    "/usermod2.py": b"print('in usermod2')",
    "/usermod3.py": b"syntax error",
    "/usermod4.mpy": b"syntax error",
    "/usermod5.py": b"print('in usermod5')",
    "/usermod6.py": b"print('in usermod6')",
}
vfs.mount(UserFS(user_files), "/userfs")

# open and read a file
f = open("/userfs/data.txt")
print(f.read())

# import files from the user filesystem
sys.path.append("/userfs")
import usermod1

# import a .py file with a syntax error (file should be closed on error)
try:
    import usermod3
except SyntaxError:
    print("SyntaxError in usermod3")

# import a .mpy file with a syntax error (file should be closed on error)
try:
    import usermod4
except ValueError:
    print("ValueError in usermod4")

# Test an import with largest buffer size
UserFile.buffer_size = 255
import usermod5

# Test an import with over-size buffer size (should be safely limited internally)
UserFile.buffer_size = 1024
import usermod6

# unmount and undo path addition
vfs.umount("/userfs")
sys.path.pop()



import time
from . import simple


class MQTTClient(simple.MQTTClient):
    DELAY = 2
    DEBUG = False

    def delay(self, i):
        time.sleep(self.DELAY)

    def log(self, in_reconnect, e):
        if self.DEBUG:
            if in_reconnect:
                print("mqtt reconnect: %r" % e)
            else:
                print("mqtt: %r" % e)

    def reconnect(self):
        i = 0
        while 1:
            try:
                return super().connect(False)
            except OSError as e:
                self.log(True, e)
                i += 1
                self.delay(i)

    def publish(self, topic, msg, retain=False, qos=0):
        while 1:
            try:
                return super().publish(topic, msg, retain, qos)
            except OSError as e:
                self.log(False, e)
            self.reconnect()

    def wait_msg(self):
        while 1:
            try:
                return super().wait_msg()
            except OSError as e:
                self.log(False, e)
            self.reconnect()

    def check_msg(self, attempts=2):
        while attempts:
            self.sock.setblocking(False)
            try:
                return super().wait_msg()
            except OSError as e:
                self.log(False, e)
            self.reconnect()
            attempts -= 1



import sys
import io
import unittest
import gc
import contextlib


TESTFN = "@test"


def run_unittest(*classes):
    suite = unittest.TestSuite()
    for c in classes:
        if isinstance(c, str):
            c = __import__(c)
            for name in dir(c):
                obj = getattr(c, name)
                if isinstance(obj, type) and issubclass(obj, unittest.TestCase):
                    suite.addTest(obj)
        else:
            suite.addTest(c)
    runner = unittest.TestRunner()
    result = runner.run(suite)


def can_symlink():
    return False


def skip_unless_symlink(test):
    """Skip decorator for tests that require functional symlink"""
    ok = can_symlink()
    msg = "Requires functional symlink implementation"
    return test if ok else unittest.skip(msg)(test)


def create_empty_file(name):
    open(name, "w").close()


@contextlib.contextmanager
def disable_gc():
    have_gc = gc.isenabled()
    gc.disable()
    try:
        yield
    finally:
        if have_gc:
            gc.enable()


def gc_collect():
    gc.collect()
    gc.collect()
    gc.collect()


@contextlib.contextmanager
def captured_output(stream_name):
    org = getattr(sys, stream_name)
    buf = io.StringIO()
    setattr(sys, stream_name, buf)
    try:
        yield buf
    finally:
        setattr(sys, stream_name, org)


def captured_stderr():
    return captured_output("stderr")


def requires_IEEE_754(f):
    return f



# Don't include default frozen modules because MCU is tight on flash space.

# Only install the sync version of the LoRa driver because this board doesn't
# have asyncio by default.
require("lora-sync")
require("lora-stm32wl5")



import hashlib
import os
import sys
import tempfile

import serial.tools.list_ports

from .transport import TransportError, stdout_write_bytes
from .transport_serial import SerialTransport


class CommandError(Exception):
    pass


def do_connect(state, args=None):
    dev = args.device[0] if args else "auto"
    do_disconnect(state)

    try:
        if dev == "list":
            # List attached devices.
            for p in sorted(serial.tools.list_ports.comports()):
                print(
                    "{} {} {:04x}:{:04x} {} {}".format(
                        p.device,
                        p.serial_number,
                        p.vid if isinstance(p.vid, int) else 0,
                        p.pid if isinstance(p.pid, int) else 0,
                        p.manufacturer,
                        p.product,
                    )
                )
            # Don't do implicit REPL command.
            state.did_action()
        elif dev == "auto":
            # Auto-detect and auto-connect to the first available USB serial port.
            for p in sorted(serial.tools.list_ports.comports()):
                if p.vid is not None and p.pid is not None:
                    try:
                        state.transport = SerialTransport(p.device, baudrate=115200)
                        return
                    except TransportError as er:
                        if not er.args[0].startswith("failed to access"):
                            raise er
            raise TransportError("no device found")
        elif dev.startswith("id:"):
            # Search for a device with the given serial number.
            serial_number = dev[len("id:") :]
            dev = None
            for p in serial.tools.list_ports.comports():
                if p.serial_number == serial_number:
                    state.transport = SerialTransport(p.device, baudrate=115200)
                    return
            raise TransportError("no device with serial number {}".format(serial_number))
        else:
            # Connect to the given device.
            if dev.startswith("port:"):
                dev = dev[len("port:") :]
            state.transport = SerialTransport(dev, baudrate=115200)
            return
    except TransportError as er:
        msg = er.args[0]
        if msg.startswith("failed to access"):
            msg += " (it may be in use by another program)"
        raise CommandError(msg)


def do_disconnect(state, _args=None):
    if not state.transport:
        return

    try:
        if state.transport.mounted:
            if not state.transport.in_raw_repl:
                state.transport.enter_raw_repl(soft_reset=False)
            state.transport.umount_local()
        if state.transport.in_raw_repl:
            state.transport.exit_raw_repl()
    except OSError:
        # Ignore any OSError exceptions when shutting down, eg:
        # - filesystem_command will close the connection if it had an error
        # - umounting will fail if serial port disappeared
        pass
    state.transport.close()
    state.transport = None
    state._auto_soft_reset = True


def show_progress_bar(size, total_size, op="copying"):
    if not sys.stdout.isatty():
        return
    verbose_size = 2048
    bar_length = 20
    if total_size < verbose_size:
        return
    elif size >= total_size:
        # Clear progress bar when copy completes
        print("\r" + " " * (13 + len(op) + bar_length) + "\r", end="")
    else:
        bar = size * bar_length // total_size
        progress = size * 100 // total_size
        print(
            "\r ... {} {:3d}% [{}{}]".format(op, progress, "#" * bar, "-" * (bar_length - bar)),
            end="",
        )


def _remote_path_join(a, *b):
    if not a:
        a = "./"
    result = a.rstrip("/")
    for x in b:
        result += "/" + x.strip("/")
    return result


def _remote_path_dirname(a):
    a = a.rsplit("/", 1)
    if len(a) == 1:
        return ""
    else:
        return a[0]


def _remote_path_basename(a):
    return a.rsplit("/", 1)[-1]


def do_filesystem_cp(state, src, dest, multiple, check_hash=False):
    if dest.startswith(":"):
        dest_no_slash = dest.rstrip("/" + os.path.sep + (os.path.altsep or ""))
        dest_exists = state.transport.fs_exists(dest_no_slash[1:])
        dest_isdir = dest_exists and state.transport.fs_isdir(dest_no_slash[1:])

        # A trailing / on dest forces it to be a directory.
        if dest != dest_no_slash:
            if not dest_isdir:
                raise CommandError("cp: destination is not a directory")
            dest = dest_no_slash
    else:
        dest_exists = os.path.exists(dest)
        dest_isdir = dest_exists and os.path.isdir(dest)

    if multiple:
        if not dest_exists:
            raise CommandError("cp: destination does not exist")
        if not dest_isdir:
            raise CommandError("cp: destination is not a directory")

    # Download the contents of source.
    try:
        if src.startswith(":"):
            data = state.transport.fs_readfile(src[1:], progress_callback=show_progress_bar)
            filename = _remote_path_basename(src[1:])
        else:
            with open(src, "rb") as f:
                data = f.read()
            filename = os.path.basename(src)
    except IsADirectoryError:
        raise CommandError("cp: -r not specified; omitting directory")

    # Write back to dest.
    if dest.startswith(":"):
        # If the destination path is just the directory, then add the source filename.
        if dest_isdir:
            dest = ":" + _remote_path_join(dest[1:], filename)

        # Skip copy if the destination file is identical.
        if check_hash:
            try:
                remote_hash = state.transport.fs_hashfile(dest[1:], "sha256")
                source_hash = hashlib.sha256(data).digest()
                # remote_hash will be None if the device doesn't support
                # hashlib.sha256 (and therefore won't match).
                if remote_hash == source_hash:
                    print("Up to date:", dest[1:])
                    return
            except OSError:
                pass

        # Write to remote.
        state.transport.fs_writefile(dest[1:], data, progress_callback=show_progress_bar)
    else:
        # If the destination path is just the directory, then add the source filename.
        if dest_isdir:
            dest = os.path.join(dest, filename)

        # Write to local file.
        with open(dest, "wb") as f:
            f.write(data)


def do_filesystem_recursive_cp(state, src, dest, multiple, check_hash):
    # Ignore trailing / on both src and dest. (Unix cp ignores them too)
    src = src.rstrip("/" + os.path.sep + (os.path.altsep if os.path.altsep else ""))
    dest = dest.rstrip("/" + os.path.sep + (os.path.altsep if os.path.altsep else ""))

    # If the destination directory exists, then we copy into it. Otherwise we
    # use the destination as the target.
    if dest.startswith(":"):
        dest_exists = state.transport.fs_exists(dest[1:])
    else:
        dest_exists = os.path.exists(dest)

    # Recursively find all files to copy from a directory.
    # `dirs` will be a list of dest split paths.
    # `files` will be a list of `(dest split path, src joined path)`.
    dirs = []
    files = []

    # For example, if src=/tmp/foo, with /tmp/foo/x.py and /tmp/foo/a/b/c.py,
    # and if the destination directory exists, then we will have:
    #   dirs = [['foo'], ['foo', 'a'], ['foo', 'a', 'b']]
    #   files = [(['foo', 'x.py'], '/tmp/foo/x.py'), (['foo', 'a', 'b', 'c.py'], '/tmp/foo/a/b/c.py')]
    # If the destination doesn't exist, then we will have:
    #   dirs = [['a'], ['a', 'b']]
    #   files = [(['x.py'], '/tmp/foo/x.py'), (['a', 'b', 'c.py'], '/tmp/foo/a/b/c.py')]

    def _list_recursive(base, src_path, dest_path, src_join_fun, src_isdir_fun, src_listdir_fun):
        src_path_joined = src_join_fun(base, *src_path)
        if src_isdir_fun(src_path_joined):
            if dest_path:
                dirs.append(dest_path)
            for entry in src_listdir_fun(src_path_joined):
                _list_recursive(
                    base,
                    src_path + [entry],
                    dest_path + [entry],
                    src_join_fun,
                    src_isdir_fun,
                    src_listdir_fun,
                )
        else:
            files.append(
                (
                    dest_path,
                    src_path_joined,
                )
            )

    if src.startswith(":"):
        src_dirname = [_remote_path_basename(src[1:])]
        dest_dirname = src_dirname if dest_exists else []
        _list_recursive(
            _remote_path_dirname(src[1:]),
            src_dirname,
            dest_dirname,
            src_join_fun=_remote_path_join,
            src_isdir_fun=state.transport.fs_isdir,
            src_listdir_fun=lambda p: [x.name for x in state.transport.fs_listdir(p)],
        )
    else:
        src_dirname = [os.path.basename(src)]
        dest_dirname = src_dirname if dest_exists else []
        _list_recursive(
            os.path.dirname(src),
            src_dirname,
            dest_dirname,
            src_join_fun=os.path.join,
            src_isdir_fun=os.path.isdir,
            src_listdir_fun=os.listdir,
        )

    # If no directories were encountered then we must have just had a file.
    if not dirs:
        return do_filesystem_cp(state, src, dest, multiple, check_hash)

    def _mkdir(a, *b):
        try:
            if a.startswith(":"):
                state.transport.fs_mkdir(_remote_path_join(a[1:], *b))
            else:
                os.mkdir(os.path.join(a, *b))
        except FileExistsError:
            pass

    # Create the destination if necessary.
    if not dest_exists:
        _mkdir(dest)

    # Create all sub-directories relative to the destination.
    for d in dirs:
        _mkdir(dest, *d)

    # Copy all files, in sorted order to help it be deterministic.
    files.sort()
    for dest_path_split, src_path_joined in files:
        if src.startswith(":"):
            src_path_joined = ":" + src_path_joined

        if dest.startswith(":"):
            dest_path_joined = ":" + _remote_path_join(dest[1:], *dest_path_split)
        else:
            dest_path_joined = os.path.join(dest, *dest_path_split)

        do_filesystem_cp(state, src_path_joined, dest_path_joined, False, check_hash)


def do_filesystem(state, args):
    state.ensure_raw_repl()
    state.did_action()

    command = args.command[0]
    paths = args.path

    if command == "cat":
        # Don't do verbose output for `cat` unless explicitly requested.
        verbose = args.verbose is True
    else:
        verbose = args.verbose is not False

    if command == "cp":
        # Note: cp requires the user to specify local/remote explicitly via
        # leading ':'.

        # The last argument must be the destination.
        if len(paths) <= 1:
            raise CommandError("cp: missing destination path")
        cp_dest = paths[-1]
        paths = paths[:-1]
    else:
        # All other commands implicitly use remote paths. Strip the
        # leading ':' if the user included them.
        paths = [path[1:] if path.startswith(":") else path for path in paths]

    # ls implicitly lists the cwd.
    if command == "ls" and not paths:
        paths = [""]

    try:
        # Handle each path sequentially.
        for path in paths:
            if verbose:
                if command == "cp":
                    print("{} {} {}".format(command, path, cp_dest))
                else:
                    print("{} :{}".format(command, path))

            if command == "cat":
                state.transport.fs_printfile(path)
            elif command == "ls":
                for result in state.transport.fs_listdir(path):
                    print(
                        "{:12} {}{}".format(
                            result.st_size, result.name, "/" if result.st_mode & 0x4000 else ""
                        )
                    )
            elif command == "mkdir":
                state.transport.fs_mkdir(path)
            elif command == "rm":
                state.transport.fs_rmfile(path)
            elif command == "rmdir":
                state.transport.fs_rmdir(path)
            elif command == "touch":
                state.transport.fs_touchfile(path)
            elif command.endswith("sum") and command[-4].isdigit():
                digest = state.transport.fs_hashfile(path, command[:-3])
                print(digest.hex())
            elif command == "cp":
                if args.recursive:
                    do_filesystem_recursive_cp(
                        state, path, cp_dest, len(paths) > 1, not args.force
                    )
                else:
                    do_filesystem_cp(state, path, cp_dest, len(paths) > 1, not args.force)
    except FileNotFoundError as er:
        raise CommandError("{}: {}: No such file or directory.".format(command, er.args[0]))
    except IsADirectoryError as er:
        raise CommandError("{}: {}: Is a directory.".format(command, er.args[0]))
    except FileExistsError as er:
        raise CommandError("{}: {}: File exists.".format(command, er.args[0]))
    except TransportError as er:
        raise CommandError("Error with transport:\n{}".format(er.args[0]))


def do_edit(state, args):
    state.ensure_raw_repl()
    state.did_action()

    if not os.getenv("EDITOR"):
        raise CommandError("edit: $EDITOR not set")
    for src in args.files:
        src = src.lstrip(":")
        dest_fd, dest = tempfile.mkstemp(suffix=os.path.basename(src))
        try:
            print("edit :%s" % (src,))
            state.transport.fs_touchfile(src)
            data = state.transport.fs_readfile(src, progress_callback=show_progress_bar)
            with open(dest_fd, "wb") as f:
                f.write(data)
            if os.system('%s "%s"' % (os.getenv("EDITOR"), dest)) == 0:
                with open(dest, "rb") as f:
                    state.transport.fs_writefile(
                        src, f.read(), progress_callback=show_progress_bar
                    )
        finally:
            os.unlink(dest)


def _do_execbuffer(state, buf, follow):
    state.ensure_raw_repl()
    state.did_action()

    try:
        state.transport.exec_raw_no_follow(buf)
        if follow:
            ret, ret_err = state.transport.follow(timeout=None, data_consumer=stdout_write_bytes)
            if ret_err:
                stdout_write_bytes(ret_err)
                sys.exit(1)
    except TransportError as er:
        raise CommandError(er.args[0])
    except KeyboardInterrupt:
        sys.exit(1)


def do_exec(state, args):
    _do_execbuffer(state, args.expr[0], args.follow)


def do_eval(state, args):
    buf = "print(" + args.expr[0] + ")"
    _do_execbuffer(state, buf, True)


def do_run(state, args):
    filename = args.path[0]
    try:
        with open(filename, "rb") as f:
            buf = f.read()
    except OSError:
        raise CommandError(f"could not read file '{filename}'")
    _do_execbuffer(state, buf, args.follow)


def do_mount(state, args):
    state.ensure_raw_repl()
    path = args.path[0]
    state.transport.mount_local(path, unsafe_links=args.unsafe_links)
    print(f"Local directory {path} is mounted at /remote")


def do_umount(state, path):
    state.ensure_raw_repl()
    state.transport.umount_local()


def do_resume(state, _args=None):
    state._auto_soft_reset = False


def do_soft_reset(state, _args=None):
    state.ensure_raw_repl(soft_reset=True)
    state.did_action()


def do_rtc(state, args):
    state.ensure_raw_repl()
    state.did_action()

    state.transport.exec("import machine")

    if args.set:
        import datetime

        now = datetime.datetime.now()
        timetuple = "({}, {}, {}, {}, {}, {}, {}, {})".format(
            now.year,
            now.month,
            now.day,
            now.weekday(),
            now.hour,
            now.minute,
            now.second,
            now.microsecond,
        )
        state.transport.exec("machine.RTC().datetime({})".format(timetuple))
    else:
        print(state.transport.eval("machine.RTC().datetime()"))



import sqlite3


conn = sqlite3.connect(":memory:")

cur = conn.cursor()
cur.execute("CREATE TABLE foo(a int)")
cur.execute("INSERT INTO foo VALUES (42)")
assert cur.lastrowid == 1
cur.execute("SELECT * FROM foo")
assert cur.fetchone() == (42,)
assert cur.fetchone() is None

cur.close()
conn.close()



# TinyS2 MicroPython Helper Library
# 2021 Seon Rozenblum - Unexpected Maker
#
# Project home:
#   https://tinys2.io
#
# 2021-Apr-10 - v0.1 - Initial implementation

# Import required libraries
from micropython import const
from machine import Pin, SPI, ADC
import machine, time

# TinyS2 Hardware Pin Assignments

# Sense Pins
VBUS_SENSE = const(21)
VBAT_SENSE = const(3)


# RGB LED Pins
RGB_DATA = const(1)
RGB_PWR = const(2)

# SPI
SPI_MOSI = const(35)
SPI_MISO = const(36)
SPI_CLK = const(37)

# I2C
I2C_SDA = const(8)
I2C_SCL = const(9)

# DAC
DAC1 = const(17)
DAC2 = const(18)


# Helper functions
def set_pixel_power(state):
    """Enable or Disable power to the onboard NeoPixel to either show colour, or to reduce power for deep sleep."""
    Pin(RGB_PWR, Pin.OUT).value(state)


def get_battery_voltage():
    """
    Returns the current battery voltage. If no battery is connected, returns 4.2V which is the charge voltage
    This is an approximation only, but useful to detect if the charge state of the battery is getting low.
    """
    adc = ADC(Pin(VBAT_SENSE))  # Assign the ADC pin to read
    measuredvbat = adc.read()  # Read the value
    measuredvbat /= 8192  # divide by 8192 as we are using the default ADC voltage range of 0-1V
    measuredvbat *= 4.2  # Multiply by 4.2V, our reference voltage
    return round(measuredvbat, 2)


def get_vbus_present():
    """Detect if VBUS (5V) power source is present"""
    return Pin(VBUS_SENSE, Pin.IN).value() == 1


# NeoPixel rainbow colour wheel
def rgb_color_wheel(wheel_pos):
    """Color wheel to allow for cycling through the rainbow of RGB colors."""
    wheel_pos = wheel_pos % 255

    if wheel_pos < 85:
        return 255 - wheel_pos * 3, 0, wheel_pos * 3
    elif wheel_pos < 170:
        wheel_pos -= 85
        return 0, wheel_pos * 3, 255 - wheel_pos * 3
    else:
        wheel_pos -= 170
        return wheel_pos * 3, 255 - wheel_pos * 3, 0


# Go into deep sleep but shut down the RGB LED first to save power
# Use this if you want lowest deep sleep current
def go_deepsleep(t):
    """Deep sleep helper that also powers down the on-board NeoPixel."""
    set_pixel_power(False)
    machine.deepsleep(t)



#!/usr/bin/env python
"""Creates the pin file for the RP2."""

import os
import re
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../../../tools"))
import boardgen

# This is NUM_BANK0_GPIOS. Pin indices are 0 to 29 (inclusive).
NUM_GPIOS = None
# Up to 32 additional extended pins (e.g. via the wifi chip or io expanders).
NUM_EXT_GPIOS = None


class Rp2Pin(boardgen.Pin):
    def __init__(self, cpu_pin_name):
        super().__init__(cpu_pin_name)
        self._afs = []

        if self.name().startswith("EXT_"):
            self._index = None
            self._ext_index = int(self.name()[8:])  # "EXT_GPIOn"
        else:
            self._index = int(self.name()[4:])  # "GPIOn"
            self._ext_index = None

    # Required by NumericPinGenerator.
    def index(self):
        return self._index

    # Use the PIN() macro defined in rp2_prefix.c for defining the pin
    # objects.
    def definition(self):
        if self._index is not None:
            return "PIN({:d}, GPIO{:d}, 0, {:d}, pin_GPIO{:d}_af)".format(
                self._index, self._index, len(self._afs), self.index()
            )
        else:
            return "PIN({:d}, EXT_GPIO{:d}, 1, 0, NULL)".format(self._ext_index, self._ext_index)

    # External pins need to be mutable (because they track the output state).
    def is_const(self):
        return self._index is not None

    # Add conditional macros only around the external pins based on how many
    # are enabled.
    def enable_macro(self):
        if self._ext_index is not None:
            return "(MICROPY_HW_PIN_EXT_COUNT > {:d})".format(self._ext_index)

    def add_af(self, af_idx, _af_name, af):
        if self._index is None:
            raise boardgen.PinGeneratorError(
                "Cannot add AF for ext pin '{:s}'".format(self.name())
            )

        # <af><unit>_<pin>
        m = re.match("([A-Z][A-Z0-9][A-Z]+)(([0-9]+)(_.*)?)?", af)
        af_fn = m.group(1)
        af_unit = int(m.group(3)) if m.group(3) is not None else 0
        if af_fn == "PIO":
            # Pins can be either PIO unit (unlike, say, I2C where a
            # pin can only be I2C0 _or_ I2C1, both sharing the same AF
            # index), so each PIO unit has a distinct AF index.
            af_fn = "{:s}{:d}".format(af_fn, af_unit)
        self._afs.append((af_idx + 1, af_fn, af_unit, af))

    # This will be called at the start of the output (after the prefix). Use
    # it to emit the af objects (via the AF() macro in rp2_prefix.c).
    def print_source(self, out_source):
        if self._index is not None:
            print(
                "const machine_pin_af_obj_t pin_GPIO{:d}_af[] = {{".format(self.index()),
                file=out_source,
            )
            for af_idx, af_fn, af_unit, af in self._afs:
                print(
                    "    AF({:d}, {:4s}, {:d}), // {:s}".format(af_idx, af_fn, af_unit, af),
                    file=out_source,
                )
            print("};", file=out_source)
            print(file=out_source)

    # rp2 cpu names must be "GPIOn" or "EXT_GPIOn".
    @staticmethod
    def validate_cpu_pin_name(cpu_pin_name):
        boardgen.Pin.validate_cpu_pin_name(cpu_pin_name)

        if cpu_pin_name.startswith("GPIO") and cpu_pin_name[4:].isnumeric():
            if not (0 <= int(cpu_pin_name[4:]) < NUM_GPIOS):
                raise boardgen.PinGeneratorError("Unknown cpu pin '{}'".format(cpu_pin_name))
        elif cpu_pin_name.startswith("EXT_GPIO") and cpu_pin_name[8:].isnumeric():
            if not (0 <= int(cpu_pin_name[8:]) < NUM_EXT_GPIOS):
                raise boardgen.PinGeneratorError("Unknown ext pin '{}'".format(cpu_pin_name))
        else:
            raise boardgen.PinGeneratorError(
                "Invalid cpu pin name '{}', must be 'GPIOn' or 'EXT_GPIOn'".format(cpu_pin_name)
            )


class Rp2PinGenerator(boardgen.NumericPinGenerator):
    def __init__(self):
        # Use custom pin type above, and also enable the --af-csv argument.
        super().__init__(
            pin_type=Rp2Pin,
            enable_af=True,
        )

    # Provided by pico-sdk.
    def cpu_table_size(self):
        return "NUM_BANK0_GPIOS"

    # Only use pre-defined cpu pins (do not let board.csv create them).
    def find_pin_by_cpu_pin_name(self, cpu_pin_name, create=True):
        return super().find_pin_by_cpu_pin_name(cpu_pin_name, create=False)

    # NumericPinGenerator doesn't include the cpu dict by default (only the
    # board dict), so add that to the output for rp2.
    def print_source(self, out_source):
        super().print_source(out_source)
        self.print_cpu_locals_dict(out_source)

    def extra_args(self, parser):
        parser.add_argument("--num-gpios", type=int)
        parser.add_argument("--num-ext-gpios", type=int)

    def load_inputs(self, out_source):
        global NUM_GPIOS
        global NUM_EXT_GPIOS

        # Needed by validate_cpu_pin_name
        NUM_GPIOS = self.args.num_gpios
        NUM_EXT_GPIOS = self.args.num_ext_gpios

        if NUM_GPIOS is None:
            raise boardgen.PinGeneratorError("Please pass num-gpios")

        if NUM_EXT_GPIOS is None:
            NUM_EXT_GPIOS = 0
        # Pre-define the pins (i.e. don't require them to be listed in pins.csv).
        for i in range(NUM_GPIOS):
            self.add_cpu_pin("GPIO{}".format(i))
        for i in range(NUM_EXT_GPIOS):
            self.add_cpu_pin("EXT_GPIO{}".format(i))

        super().load_inputs(out_source)


if __name__ == "__main__":
    Rp2PinGenerator().main()



# This directory contains all aioble code, but the manifest itself just
# forwards to the component manifests, which themselves reference the actual
# code. This allows (for development purposes) all the files to live in the
# one directory.

metadata(version="0.6.0")

# Default installation gives you everything. Install the individual
# components (or a combination of them) if you want a more minimal install.
require("aioble-peripheral")
require("aioble-server")
require("aioble-central")
require("aioble-client")
require("aioble-l2cap")
require("aioble-security")



# MicroPython LoRa SX126x (SX1261, SX1262) driver
# MIT license; Copyright (c) 2023 Angus Gratton
#
# LoRa is a registered trademark or service mark of Semtech Corporation or its affiliates.
#
# In comments, abbreviation "DS" = Semtech SX1261/62 Datasheet Rev 2.1 (December 2021)
import struct
import time
from micropython import const
from machine import Pin
from .modem import BaseModem, ConfigError, RxPacket, _clamp, _flag

# Set _DEBUG to const(True) to print all SPI commands sent to the device, and all responses,
# plus a few additional pieces of information.
_DEBUG = const(False)

_REG_RXGAINCR = const(0x8AC)  # Reset value 0x94
_REG_LSYNCRH = const(0x740)
_REG_LSYNCRL = const(0x741)

_CMD_CFG_DIO_IRQ = const(0x08)  # args: IrqMask, Irq1Mask, Irq2Mask,. Irq3Mask
_CMD_CLR_ERRORS = const(0x07)
_CMD_CLR_IRQ_STATUS = const(0x02)  # no args
_CMD_GET_ERROR = const(0x17)
_CMD_GET_IRQ_STATUS = const(0x12)  # args: (r) Status, IrqStatus
_CMD_GET_RX_BUFFER_STATUS = const(0x13)  # args: (r) Status, RxPayloadLength, RxBufferPointer
# NOTE: _CMD_GET_STATUS seems to have an issue, see _get_status() function below.
_CMD_GET_STATUS = const(0xC0)  # args: (r) Status
_CMD_GET_PACKET_STATUS = const(0x14)
_CMD_READ_REGISTER = const(0x1D)  # args: addr (2b), status, (r) Data0 ... DataN
_CMD_READ_BUFFER = const(0x1E)  # args: Offset, (r) Status, Data0 ... DataN
_CMD_SET_BUFFER_BASE_ADDRESS = const(0x8F)  # args: TxBaseAddr, RxBaseAddr
_CMD_SET_MODULATION_PARAMS = const(0x8B)  # args (LoRa): Sf, Bw, Cr, Ldro
_CMD_SET_PACKET_PARAMS = const(
    0x8C
)  # args (LoRa): PbLength, HeaderType, PayloadLength, CrcType, InvertIQ
_CMD_SET_PACKET_TYPE = const(0x8A)  # args: PktType
_CMD_SET_PA_CONFIG = const(0x95)  # args: PaDutyCycle, HpMax, HpSel, 0x01
_CMD_SET_RF_FREQUENCY = const(0x86)  # args: RfFreg
_CMD_SET_RX = const(0x82)  # args: Timeout
_CMD_SET_SLEEP = const(0x84)  # args: SleepCfg
_CMD_SET_STANDBY = const(0x80)  # args: StandbyCfg
_CMD_SET_DIO3_AS_TCXO_CTRL = const(0x97)  # args: Trim, Timeout (3b)
_CMD_SET_DIO2_AS_RF_SWITCH_CTRL = const(0x9D)
_CMD_SET_TX = const(0x83)  # args: Timeout
_CMD_SET_TX_PARAMS = const(0x8E)  # args: Power, RampTime
_CMD_WRITE_BUFFER = const(0x0E)  # args: Offset, Data0 ... DataN
_CMD_WRITE_REGISTER = const(0x0D)  # args: Addr, Data0 ... Data N

_CMD_CALIBRATE = const(0x89)
_CMD_CALIBRATE_IMAGE = const(0x98)

_STATUS_MODE_MASK = const(0x7 << 4)
_STATUS_MODE_SHIFT = const(4)
_STATUS_MODE_STANDBY_RC = const(0x2)
_STATUS_MODE_STANDBY_HSE32 = const(0x3)
_STATUS_MODE_FS = const(0x4)
_STATUS_MODE_RX = const(0x5)
_STATUS_MODE_TX = const(0x6)

_STATUS_CMD_MASK = const(0x6)  # bits 1-3, bit 0 is reserved
_STATUS_CMD_SHIFT = const(1)
_STATUS_CMD_DATA_AVAIL = const(0x2)
_STATUS_CMD_TIMEOUT = const(0x3)
_STATUS_CMD_ERROR = const(0x4)
_STATUS_CMD_EXEC_FAIL = const(0x5)
_STATUS_CMD_TX_COMPLETE = const(0x6)

_CFG_SF_MIN = const(6)  # inclusive
_CFG_SF_MAX = const(12)  # inclusive

_IRQ_TX_DONE = const(1 << 0)
_IRQ_RX_DONE = const(1 << 1)
_IRQ_PREAMBLE_DETECTED = const(1 << 2)
_IRQ_SYNC_DETECTED = const(1 << 3)
_IRQ_HEADER_VALID = const(1 << 4)
_IRQ_HEADER_ERR = const(1 << 5)
_IRQ_CRC_ERR = const(1 << 6)
_IRQ_CAD_DONE = const(1 << 7)
_IRQ_CAD_DETECTED = const(1 << 8)
_IRQ_TIMEOUT = const(1 << 9)

# Register values
_REG_IQ_POLARITY_SETUP = const(0x0736)
_REG_RX_GAIN = const(0x08AC)
_REG_RTC_CTRL = const(0x0902)  # DS 15.3 has a typo on this value! Confirmed from Semtech driver
_REG_EVT_CLR = const(0x0944)
_REG_EVT_CLR_MASK = const(0x02)

# IRQs the driver cares about when receiving
_IRQ_DRIVER_RX_MASK = const(_IRQ_RX_DONE | _IRQ_TIMEOUT | _IRQ_CRC_ERR | _IRQ_HEADER_ERR)


# Except when entering/waking from sleep, typical busy period <105us (ref RM0453 Table 33)
#
# However, if dio3_tcxo_start_time_us is set then can take a longer
# time to become valid, so a field is set on the modem object with the full timeout.
#
# In any case, timeouts here are to catch broken/bad hardware or massive driver
# bugs rather than commonplace issues.
#
_CMD_BUSY_TIMEOUT_BASE_US = const(7000)

# Datasheet says 3.5ms needed to run a full Calibrate command (all blocks),
# however testing shows it can be as much as as 18ms.
_CALIBRATE_TYPICAL_TIME_US = const(3500)
_CALIBRATE_TIMEOUT_US = const(30000)

# Magic value used by SetRx command to indicate a continuous receive
_CONTINUOUS_TIMEOUT_VAL = const(0xFFFFFF)


class _SX126x(BaseModem):
    # common IRQ masks used by the base class functions
    _IRQ_RX_COMPLETE = _IRQ_RX_DONE | _IRQ_TIMEOUT
    _IRQ_TX_COMPLETE = _IRQ_TX_DONE

    # Common base class for SX1261, SX1262 and (pending) STM32WL55. These are all basically
    # the same except for which PA ranges are supported
    #
    # Don't construct this directly, construct lora.SX1261, lora.SX1262, lora.AsyncSX1261
    # or lora.AsyncSX1262
    def __init__(
        self,
        spi,
        cs,
        busy,
        dio1,
        dio2_rf_sw,
        dio3_tcxo_millivolts,
        dio3_tcxo_start_time_us,
        reset,
        lora_cfg,
        ant_sw,
    ):
        super().__init__(ant_sw)

        self._spi = spi
        self._cs = cs
        self._busy = busy
        self._sleep = True  # assume the radio is in sleep mode to start, will wake on _cmd
        self._dio1 = dio1

        if hasattr(busy, "init"):
            busy.init(Pin.IN)
        if hasattr(cs, "init"):
            cs.init(Pin.OUT, value=1)
        if hasattr(dio1, "init"):
            dio1.init(Pin.IN)

        self._busy_timeout = _CMD_BUSY_TIMEOUT_BASE_US + (
            dio3_tcxo_start_time_us if dio3_tcxo_millivolts else 0
        )

        self._buf_view = memoryview(bytearray(9))  # shared buffer for commands

        # These settings are kept in the object (as can't read them back from the modem)
        self._output_power = 14
        self._bw = 125
        assert self._bw_hz == 125000  # This field is set in base class, must match self._bw

        # RampTime register value
        # 0x02 is 40us, default value appears undocumented but this is the SX1276 default
        self._ramp_val = 0x02

        # Configure the SX126x at least once after reset
        self._configured = False

        if reset:
            # If the caller supplies a reset pin argument, reset the radio
            reset.init(Pin.OUT, value=0)
            time.sleep_ms(1)
            reset(1)
            time.sleep_ms(5)
        else:
            # Otherwise, at least put the radio to a known state
            self._cmd("BB", _CMD_SET_STANDBY, 0)  # STDBY_RC mode, not ready for TCXO yet

        status = self._get_status()
        if (status[0] != _STATUS_MODE_STANDBY_RC and status[0] != _STATUS_MODE_STANDBY_HSE32) or (
            status[1] > 1
        ):
            # This check exists to determine that the SPI settings and modem
            # selection are correct. Otherwise it's possible for the driver to
            # run for quite some time before it detects an invalid response.
            raise RuntimeError("Invalid initial status {}.".format(status))

        if dio2_rf_sw:
            self._cmd("BB", _CMD_SET_DIO2_AS_RF_SWITCH_CTRL, 1)

        if dio3_tcxo_millivolts:
            # Enable TCXO power via DIO3, if enabled
            #
            # timeout register is set in units of 15.625us each, use integer math
            # to calculate and round up:
            timeout = (dio3_tcxo_start_time_us * 1000 + 15624) // 15625
            if timeout < 0 or timeout > 1 << 24:
                raise ValueError("{} out of range".format("dio3_tcxo_start_time_us"))
            if dio3_tcxo_millivolts < 1600 or dio3_tcxo_millivolts > 3300:
                raise ValueError("{} out of range".format("dio3_tcxo_millivolts"))
            dv = dio3_tcxo_millivolts // 100  # 16 to 33
            tcxo_trim_lookup = (
                16,
                17,
                18,
                22,
                24,
                27,
                30,
                33,
            )  # DS Table 13-35
            while dv not in tcxo_trim_lookup:
                dv -= 1
            reg_tcxo_trim = tcxo_trim_lookup.index(dv)

            self._cmd(">BI", _CMD_SET_DIO3_AS_TCXO_CTRL, (reg_tcxo_trim << 24) + timeout)
            time.sleep_ms(15)
            # As per DS 13.3.6 SetDIO3AsTCXOCtrl, should expect error
            # value 0x20 "XOSC_START_ERR" to be flagged as XOSC has only just
            # started now. So clear it.
            self._clear_errors()

            self._check_error()

        # If DIO1 is set, mask in just the IRQs that the driver may need to be
        # interrupted by. This is important because otherwise an unrelated IRQ
        # can trigger the ISR and may not be reset by the driver, leaving DIO1 high.
        #
        # If DIO1 is not set, all IRQs can stay masked which is the power-on state.
        if dio1:
            # Note: we set both Irq mask and DIO1 mask to the same value, which is redundant
            # (one could be 0xFFFF) but may save a few bytes of bytecode.
            self._cmd(
                ">BHHHH",
                _CMD_CFG_DIO_IRQ,
                (_IRQ_RX_DONE | _IRQ_TX_DONE | _IRQ_TIMEOUT),  # IRQ mask
                (_IRQ_RX_DONE | _IRQ_TX_DONE | _IRQ_TIMEOUT),  # DIO1 mask
                0x0,  # DIO2Mask, not used
                0x0,  # DIO3Mask, not used
            )
            dio1.irq(self._radio_isr, Pin.IRQ_RISING)

        self._clear_irq()

        self._cmd("BB", _CMD_SET_PACKET_TYPE, 1)  # LoRa

        if lora_cfg:
            self.configure(lora_cfg)

    def sleep(self, warm_start=True):
        # Put the modem into sleep mode. Driver will wake the modem automatically the next
        # time an operation starts, or call standby() to wake it manually.
        #
        # If the warm_start parameter is False (non-default) then the modem will
        # lose all settings on wake. The only way to use this parameter value is
        # to destroy this modem object after calling it, and then instantiate a new
        # modem object on wake.
        #
        self._check_error()  # check errors before going to sleep because we clear on wake
        self.standby()  # save some code size, this clears the driver's rx/tx state
        self._cmd("BB", _CMD_SET_SLEEP, _flag(1 << 2, warm_start))
        self._sleep = True

    def _standby(self):
        # Send the command for standby mode.
        #
        # **Don't call this function directly, call standby() instead.**
        #
        # (This private version doesn't update the driver's internal state.)
        self._cmd("BB", _CMD_SET_STANDBY, 1)  # STDBY_XOSC mode
        self._clear_irq()  # clear IRQs in case we just cancelled a send or receive

    def is_idle(self):
        # Returns True if the modem is idle (either in standby or in sleep).
        #
        # Note this function can return True in the case where the modem has temporarily gone to
        # standby but there's a receive configured in software that will resume receiving the next
        # time poll_recv() or poll_send() is called.
        if self._sleep:
            return True  # getting status wakes from sleep
        mode, _ = self._get_status()
        return mode in (_STATUS_MODE_STANDBY_HSE32, _STATUS_MODE_STANDBY_RC)

    def _wakeup(self):
        # Wake the modem from sleep. This is called automatically the first
        # time a modem command is sent after sleep() was called to put the modem to
        # sleep.
        #
        # To manually wake the modem without initiating a new operation, call standby().
        self._cs(0)
        time.sleep_us(20)
        self._cs(1)
        self._sleep = False
        self._clear_errors()  # Clear "XOSC failed to start" which will reappear at this time
        self._check_error()  # raise an exception if any other error appears

    def _decode_status(self, raw_status, check_errors=True):
        # split the raw status, which often has reserved bits set, into the mode value
        # and the command status value
        mode = (raw_status & _STATUS_MODE_MASK) >> _STATUS_MODE_SHIFT
        cmd = (raw_status & _STATUS_CMD_MASK) >> _STATUS_CMD_SHIFT
        if check_errors and cmd in (_STATUS_CMD_EXEC_FAIL, _STATUS_CMD_ERROR):
            raise RuntimeError("Status {},{} indicates command error".format(mode, cmd))
        return (mode, cmd)

    def _get_status(self):
        # Issue the GetStatus command and return the decoded status of (mode
        # value, command status)
        #
        # Due to what appears to be a silicon bug, we send GetIrqStatus here
        # instead of GetStatus. It seems that there is some specific sequence
        # where sending command GetStatus to the chip immediately after SetRX
        # (mode 5) will trip it it into an endless TX (mode 6) for no apparent
        # reason!
        #
        # It doesn't seem to be timing dependent, all that's needed is that
        # ordering (and the modem works fine otherwise).
        #
        # As a workaround we send the GetIrqStatus command and read an extra two
        # bytes that are then ignored...
        res = self._cmd("B", _CMD_GET_IRQ_STATUS, n_read=3)[0]
        return self._decode_status(res)

    def _check_error(self):
        # Raise a RuntimeError if the radio has reported an error state.
        #
        # Return the decoded status, otherwise.
        res = self._cmd("B", _CMD_GET_ERROR, n_read=3)
        status = self._decode_status(res[0], False)
        op_error = (res[1] << 8) + res[2]
        if op_error != 0:
            raise RuntimeError("Internal radio Status {} OpError {:#x}".format(status, op_error))
        self._decode_status(res[0])  # raise an exception here if status shows an error
        return status

    def _clear_errors(self):
        # Clear any errors flagged in the modem
        self._cmd(">BH", _CMD_CLR_ERRORS, 0)

    def _clear_irq(self, clear_bits=0xFFFF):
        # Clear IRQs flagged in the modem
        #
        # By default, clears all IRQ bits. Otherwise, argument is the mask of bits to clear.
        self._cmd(">BH", _CMD_CLR_IRQ_STATUS, clear_bits)
        self._last_irq = None

    def _set_tx_ant(self, tx_ant):
        # Only STM32WL55 allows switching tx_ant from LP to HP
        raise ConfigError("tx_ant")

    def _symbol_offsets(self):
        # Called from BaseModem.get_time_on_air_us().
        #
        # This function provides a way to implement the different SF5 and SF6 in SX126x,
        # by returning two offsets: one for the overall number of symbols, and one for the
        # number of bits used to calculate the symbol length of the payload.
        return (2, -8) if self._sf in (5, 6) else (0, 0)

    def configure(self, lora_cfg):
        if self._rx is not False:
            raise RuntimeError("Receiving")

        if "preamble_len" in lora_cfg:
            self._preamble_len = lora_cfg["preamble_len"]

        self._invert_iq = (
            lora_cfg.get("invert_iq_rx", self._invert_iq[0]),
            lora_cfg.get("invert_iq_tx", self._invert_iq[1]),
            self._invert_iq[2],
        )

        if "freq_khz" in lora_cfg:
            self._rf_freq_hz = int(lora_cfg["freq_khz"] * 1000)
            rffreq = (
                self._rf_freq_hz << 25
            ) // 32_000_000  # RF-PLL frequency = 32e^6 * RFFreq / 2^25
            if not rffreq:
                raise ConfigError("freq_khz")  # set to a value too low
            self._cmd(">BI", _CMD_SET_RF_FREQUENCY, rffreq)

        if "syncword" in lora_cfg:
            syncword = lora_cfg["syncword"]
            if syncword < 0x100:
                # "Translation from SX127x to SX126x : 0xYZ -> 0xY4Z4 :
                # if you do not set the two 4 you might lose sensitivity"
                # see
                # https://www.thethingsnetwork.org/forum/t/should-private-lorawan-networks-use-a-different-sync-word/34496/15
                syncword = 0x0404 + ((syncword & 0x0F) << 4) + ((syncword & 0xF0) << 8)
            self._cmd(">BHH", _CMD_WRITE_REGISTER, _REG_LSYNCRH, syncword)

        if not self._configured or any(
            key in lora_cfg for key in ("output_power", "pa_ramp_us", "tx_ant")
        ):
            pa_config_args, self._output_power = self._get_pa_tx_params(
                lora_cfg.get("output_power", self._output_power), lora_cfg.get("tx_ant", None)
            )
            self._cmd("BBBBB", _CMD_SET_PA_CONFIG, *pa_config_args)

            if "pa_ramp_us" in lora_cfg:
                self._ramp_val = self._get_pa_ramp_val(
                    lora_cfg, [10, 20, 40, 80, 200, 800, 1700, 3400]
                )

            self._cmd("BBB", _CMD_SET_TX_PARAMS, self._output_power, self._ramp_val)

        if not self._configured or any(key in lora_cfg for key in ("sf", "bw", "coding_rate")):
            if "sf" in lora_cfg:
                self._sf = lora_cfg["sf"]
                if self._sf < _CFG_SF_MIN or self._sf > _CFG_SF_MAX:
                    raise ConfigError("sf")

            if "bw" in lora_cfg:
                self._bw = lora_cfg["bw"]

            if "coding_rate" in lora_cfg:
                self._coding_rate = lora_cfg["coding_rate"]
                if self._coding_rate < 4 or self._coding_rate > 8:  # 4/4 through 4/8, linearly
                    raise ConfigError("coding_rate")

            bw_val, self._bw_hz = {
                "7.8": (0x00, 7800),
                "10.4": (0x08, 10400),
                "15.6": (0x01, 15600),
                "20.8": (0x09, 20800),
                "31.25": (0x02, 31250),
                "41.7": (0x0A, 41700),
                "62.5": (0x03, 62500),
                "125": (0x04, 125000),
                "250": (0x05, 250000),
                "500": (0x06, 500000),
            }[str(self._bw)]

            self._cmd(
                "BBBBB",
                _CMD_SET_MODULATION_PARAMS,
                self._sf,
                bw_val,
                self._coding_rate - 4,  # 4/4=0, 4/5=1, etc
                self._get_ldr_en(),  # Note: BaseModem.get_n_symbols_x4() depends on this logic
            )

        if "rx_boost" in lora_cfg:
            # See DS Table 9-3 "Rx Gain Configuration"
            self._reg_write(_REG_RX_GAIN, 0x96 if lora_cfg["rx_boost"] else 0x94)

        self._check_error()
        self._configured = True

    def _invert_workaround(self, enable):
        # Apply workaround for DS 15.4 Optimizing the Inverted IQ Operation
        if self._invert_iq[2] != enable:
            val = self._read_read(_REG_IQ_POLARITY_SETUP)
            val = (val & ~4) | _flag(4, enable)
            self._reg_write(_REG_IQ_POLARITY_SETUP, val)
            self._invert_iq[2] = enable

    def _get_irq(self):
        # Get currently set IRQ bits.
        irq_status = self._cmd("B", _CMD_GET_IRQ_STATUS, n_read=3)
        status = self._decode_status(irq_status[0])
        flags = (irq_status[1] << 8) + irq_status[2]
        if _DEBUG:
            print("Status {} flags {:#x}".format(status, flags))
        return flags

    def calibrate(self):
        # Send the Calibrate command to the radio to calibrate RC oscillators, PLL and ADC.
        #
        # See DS 13.1.12 Calibrate Function

        # calibParam 0xFE means to calibrate all blocks.
        self._cmd("BB", _CMD_CALIBRATE, 0xFE)

        time.sleep_us(_CALIBRATE_TYPICAL_TIME_US)

        # a falling edge of BUSY indicates calibration is done
        self._wait_not_busy(_CALIBRATE_TIMEOUT_US)

    def calibrate_image(self):
        # Send the CalibrateImage command to the modem to improve reception in
        # the currently configured frequency band.
        #
        # See DS 9.2.1 Image Calibration for Specified Frequency Bands
        # and 13.1.13 CalibrateImage

        mhz = self._rf_freq_hz // 1_000_000
        if 430 <= mhz <= 440:
            args = 0x6B6F
        elif 470 <= mhz <= 510:
            args = 0x7581
        elif 779 <= mhz <= 787:
            args = 0xC1C5
        elif 863 <= mhz <= 870:
            args = 0xD7DB
        elif 902 <= mhz <= 928:
            args = 0xE1E9
        else:
            # DS says "Contact your Semtech representative for the other optimal
            # calibration settings outside of the given frequency bands"
            raise ValueError

        self._cmd(">BH", _CMD_CALIBRATE_IMAGE, args)

        # Can't find anythign in Datasheet about how long image calibration
        # takes or exactly how it signals completion. Assuming it will be
        # similar to _CMD_CALIBRATE.
        self._wait_not_busy(_CALIBRATE_TIMEOUT_US)

    def start_recv(self, timeout_ms=None, continuous=False, rx_length=0xFF):
        # Start receiving.
        #
        # Part of common low-level modem API, see README.md for usage.
        super().start_recv(timeout_ms, continuous, rx_length)  # sets _rx

        if self._tx:
            # Send is in progress and has priority, _check_recv() will start recv
            # once send finishes (caller needs to call poll_send() for this to happen.)
            if _DEBUG:
                print("Delaying receive until send completes")
            return self._dio1

        # Put the modem in a known state. It's possible a different
        # receive was in progress, this prevent anything changing while
        # we set up the new receive
        self._standby()  # calling private version to keep driver state as-is

        # Allocate the full FIFO for RX
        self._cmd("BBB", _CMD_SET_BUFFER_BASE_ADDRESS, 0xFF, 0x0)

        self._cmd(
            ">BHBBBB",
            _CMD_SET_PACKET_PARAMS,
            self._preamble_len,
            self._implicit_header,
            rx_length,  # PayloadLength, only used in implicit header mode
            self._crc_en,  # CRCType, only used in implicit header mode
            self._invert_iq[0],  # InvertIQ
        )
        self._invert_workaround(self._invert_iq[0])

        if continuous:
            timeout = _CONTINUOUS_TIMEOUT_VAL
        elif timeout_ms is not None:
            timeout = max(1, timeout_ms * 64)  # units of 15.625us
        else:
            timeout = 0  # Single receive mode, no timeout

        self._cmd(">BBH", _CMD_SET_RX, timeout >> 16, timeout)  # 24 bits

        return self._dio1

    def poll_recv(self, rx_packet=None):
        old_rx = self._rx
        rx = super().poll_recv(rx_packet)

        if rx is not True and old_rx is not False and isinstance(old_rx, int):
            # Receiving has just stopped, and a timeout was previously set.
            #
            # Workaround for errata DS 15.3 "Implicit Header Mode Timeout Behaviour",
            # which recommends to add the following after "ANY Rx with Timeout active sequence"
            self._reg_write(_REG_RTC_CTRL, 0x00)
            self._reg_write(_REG_EVT_CLR, self._reg_read(_REG_EVT_CLR) | _REG_EVT_CLR_MASK)

        return rx

    def _rx_flags_success(self, flags):
        # Returns True if IRQ flags indicate successful receive.
        # Specifically, from the bits in _IRQ_DRIVER_RX_MASK:
        # - _IRQ_RX_DONE must be set
        # - _IRQ_TIMEOUT must not be set
        # - _IRQ_CRC_ERR must not be set
        # - _IRQ_HEADER_ERR must not be set
        #
        # (Note: this is a function because the result for SX1276 depends on
        # current config, but the result is constant here.)
        return flags & _IRQ_DRIVER_RX_MASK == _IRQ_RX_DONE

    def _read_packet(self, rx_packet, flags):
        # Private function to read received packet (RxPacket object) from the
        # modem, if there is one.
        #
        # Called from poll_recv() function, which has already checked the IRQ flags
        # and verified a valid receive happened.

        ticks_ms = self._get_last_irq()

        res = self._cmd("B", _CMD_GET_RX_BUFFER_STATUS, n_read=3)
        rx_payload_len = res[1]
        rx_buffer_ptr = res[2]  # should be 0

        if rx_packet is None or len(rx_packet) != rx_payload_len:
            rx_packet = RxPacket(rx_payload_len)

        self._cmd("BB", _CMD_READ_BUFFER, rx_buffer_ptr, n_read=1, read_buf=rx_packet)

        pkt_status = self._cmd("B", _CMD_GET_PACKET_STATUS, n_read=4)

        rx_packet.ticks_ms = ticks_ms
        rx_packet.snr = pkt_status[2]  # SNR, units: dB *4
        rx_packet.rssi = 0 - pkt_status[1] // 2  # RSSI, units: dBm
        rx_packet.crc_error = (flags & _IRQ_CRC_ERR) != 0

        return rx_packet

    def prepare_send(self, packet):
        # Prepare modem to start sending. Should be followed by a call to start_send()
        #
        # Part of common low-level modem API, see README.md for usage.
        if len(packet) > 255:
            raise ConfigError("packet too long")

        # Put the modem in a known state. Any current receive is suspended at this point,
        # but calling _check_recv() will resume it later.
        self._standby()  # calling private version to keep driver state as-is

        self._check_error()

        # Set the board antenna for correct TX mode
        if self._ant_sw:
            self._ant_sw.tx(self._tx_hp())

        self._last_irq = None

        self._cmd(
            ">BHBBBB",
            _CMD_SET_PACKET_PARAMS,
            self._preamble_len,
            self._implicit_header,
            len(packet),
            self._crc_en,
            self._invert_iq[1],  # _invert_iq_tx
        )
        self._invert_workaround(self._invert_iq[1])

        # Allocate the full FIFO for TX
        self._cmd("BBB", _CMD_SET_BUFFER_BASE_ADDRESS, 0x0, 0xFF)
        self._cmd("BB", _CMD_WRITE_BUFFER, 0x0, write_buf=packet)

        # Workaround for DS 15.1 Modulation Quality with 500 kHZ LoRa Bandwidth
        # ... apparently this needs to be done "*before each packet transmission*"
        if self._bw_hz == 500_000:
            self._reg_write(0x0889, self._reg_read(0x0889) & 0xFB)
        else:
            self._reg_write(0x0889, self._reg_read(0x0889) | 0x04)

    def start_send(self):
        # Actually start a send that was loaded by calling prepare_send().
        #
        # This is split into a separate function to allow more precise timing.
        #
        # The driver doesn't verify the caller has done the right thing here, the
        # modem will no doubt do something weird if prepare_send() was not called!
        #
        # Part of common low-level modem API, see README.md for usage.

        # Currently we don't pass any TX timeout argument to the modem1,
        # which the datasheet ominously offers as "security" for the Host MCU if
        # the send doesn't start for some reason.

        self._cmd("BBBB", _CMD_SET_TX, 0x0, 0x0, 0x0)

        if _DEBUG:
            print("status {}".format(self._get_status()))
        self._check_error()

        self._tx = True

        return self._dio1

    def _wait_not_busy(self, timeout_us):
        # Wait until the radio de-asserts the busy line
        start = time.ticks_us()
        ticks_diff = 0
        while self._busy():
            ticks_diff = time.ticks_diff(time.ticks_us(), start)
            if ticks_diff > timeout_us:
                raise RuntimeError("BUSY timeout", timeout_us)
            time.sleep_us(1)
        if _DEBUG and ticks_diff > 105:
            # By default, debug log any busy time that takes longer than the
            # datasheet-promised Typical 105us (this happens when starting the 32MHz oscillator,
            # if it's turned on and off by the modem, and maybe other times.)
            print(f"BUSY {ticks_diff}us")

    def _cmd(self, fmt, *write_args, n_read=0, write_buf=None, read_buf=None):
        # Execute an SX1262 command
        # fmt - Format string suitable for use with struct.pack. First item should be 'B' and
        # corresponds to the command opcode.
        # write_args - Arguments suitable for struct.pack using fmt. First argument should be a
        # command opcode byte.
        #
        # Optional arguments:
        # write_buf - Extra buffer to write from (for FIFO writes). Mutually exclusive with n_read
        # or read_buf.
        # n_read - Number of result bytes to read back at end
        # read_buf - Extra buffer to read into (for FIFO reads)
        #
        # Returns None if n_read==0, otherwise a memoryview of length n_read which points into a
        # shared buffer (buffer will be clobbered on next call to _cmd!)
        if self._sleep:
            self._wakeup()

        # Ensure "busy" from previously issued command has de-asserted. Usually this will
        # have happened well before _cmd() is called again.
        self._wait_not_busy(self._busy_timeout)

        # Pack write_args into slice of _buf_view memoryview of correct length
        wrlen = struct.calcsize(fmt)
        assert n_read + wrlen <= len(self._buf_view)  # if this fails, make _buf bigger!
        struct.pack_into(fmt, self._buf_view, 0, *write_args)
        buf = self._buf_view[: (wrlen + n_read)]

        if _DEBUG:
            print(">>> {}".format(buf[:wrlen].hex()))
            if write_buf:
                print(">>> {}".format(write_buf.hex()))
        self._cs(0)
        self._spi.write_readinto(buf, buf)
        if write_buf:
            self._spi.write(write_buf)  # Used by _CMD_WRITE_BUFFER only
        if read_buf:
            self._spi.readinto(read_buf, 0xFF)  # Used by _CMD_READ_BUFFER only
        self._cs(1)

        if n_read > 0:
            res = self._buf_view[wrlen : (wrlen + n_read)]  # noqa: E203
            if _DEBUG:
                print("<<< {}".format(res.hex()))
            return res

    def _reg_read(self, addr):
        return self._cmd(">BHB", _CMD_READ_REGISTER, addr, 0, n_read=1)[0]

    def _reg_write(self, addr, val):
        return self._cmd(">BHB", _CMD_WRITE_REGISTER, addr, val & 0xFF)


class _SX1262(_SX126x):
    # Don't construct this directly, construct lora.SX1262 or lora.AsyncSX1262
    def __init__(
        self,
        spi,
        cs,
        busy,
        dio1=None,
        dio2_rf_sw=True,
        dio3_tcxo_millivolts=None,
        dio3_tcxo_start_time_us=1000,
        reset=None,
        lora_cfg=None,
        ant_sw=None,
    ):
        super().__init__(
            spi,
            cs,
            busy,
            dio1,
            dio2_rf_sw,
            dio3_tcxo_millivolts,
            dio3_tcxo_start_time_us,
            reset,
            lora_cfg,
            ant_sw,
        )

        # Apply workaround for DS 15.2 "Better Resistance of the SX1262 Tx to Antenna Mismatch
        self._reg_write(0x8D8, self._reg_read(0x8D8) | 0x1E)

    def _tx_hp(self):
        # SX1262 has High Power only (deviceSel==0)
        return True

    def _get_pa_tx_params(self, output_power, tx_ant):
        # Given an output power level in dB, return a 2-tuple:
        # - First item is the 3 arguments for SetPaConfig command
        # - Second item is the power level argument value for SetTxParams command.
        #
        # DS 13.1.14.1 "PA Optimal Settings" gives optimally efficient
        # values for output power +22, +20, +17, +14 dBm and "these changes make
        # the use of nominal power either sub-optimal or unachievable" (hence it
        # recommends setting +22dBm nominal TX Power for all these).
        #
        # However the modem supports output power as low as -9dBm, and there's
        # no explanation in the datasheet of how to best set other output power
        # levels.
        #
        # Semtech's own driver (sx126x.c in LoRaMac-node) only ever executes
        # SetPaConfig with the values shown in the datasheet for +22dBm, and
        # then executes SetTxParams with power set to the nominal value in
        # dBm.
        #
        # Try for best of both worlds here: If the caller requests an "Optimal"
        # value, use the datasheet values. Otherwise set nominal power only as
        # per Semtech's driver.
        output_power = int(_clamp(output_power, -9, 22))

        DEFAULT = (0x4, 0x7, 0x0, 0x1)
        OPTIMAL = {
            22: (DEFAULT, 22),
            20: ((0x3, 0x5, 0x0, 0x1), 22),
            17: ((0x2, 0x3, 0x0, 0x1), 22),
            14: ((0x2, 0x2, 0x0, 0x1), 22),
        }
        if output_power in OPTIMAL:
            # Datasheet optimal values
            return OPTIMAL[output_power]
        else:
            # Nominal values, as per Semtech driver
            return (DEFAULT, output_power & 0xFF)


class _SX1261(_SX126x):
    # Don't construct this directly, construct lora.SX1261, or lora.AsyncSX1261
    def __init__(
        self,
        spi,
        cs,
        busy,
        dio1=None,
        dio2_rf_sw=True,
        dio3_tcxo_millivolts=None,
        dio3_tcxo_start_time_us=1000,
        reset=None,
        lora_cfg=None,
        ant_sw=None,
    ):
        super().__init__(
            spi,
            cs,
            busy,
            dio1,
            dio2_rf_sw,
            dio3_tcxo_millivolts,
            dio3_tcxo_start_time_us,
            reset,
            lora_cfg,
            ant_sw,
        )

    def _tx_hp(self):
        # SX1261 has Low Power only (deviceSel==1)
        return False

    def _get_pa_tx_params(self, output_power, tx_ant):
        # Given an output power level in dB, return a 2-tuple:
        # - First item is the 3 arguments for SetPaConfig command
        # - Second item is the power level argument value for SetTxParams command.
        #
        # As noted above for SX1262, DS 13.1.14.1 "PA Optimal Settings"
        # gives optimally efficient values for output power +15, +14, +10 dBm
        # but nothing specific to the other power levels (down to -17dBm).
        #
        # Therefore do the same as for SX1262 to set optimal values if known, nominal otherwise.
        output_power = _clamp(int(output_power), -17, 15)

        DEFAULT = (0x4, 0x0, 0x1, 0x1)
        OPTIMAL = {
            15: ((0x06, 0x0, 0x1, 0x1), 14),
            14: (DEFAULT, 14),
            10: ((0x1, 0x0, 0x1, 0x1), 13),
        }

        if output_power == 15 and self._rf_freq_hz < 400_000_000:
            # DS 13.1.14.1 has Note that PaDutyCycle is limited to 0x4 below 400MHz,
            # so disallow the 15dBm optimal setting.
            output_power = 14

        if output_power in OPTIMAL:
            # Datasheet optimal values
            return OPTIMAL[output_power]
        else:
            # Nominal values, as per Semtech driver
            return (DEFAULT, output_power & 0xFF)


# Define the actual modem classes that use the SyncModem & AsyncModem "mixin-like" classes
# to create sync and async variants.

try:
    from .sync_modem import SyncModem

    class SX1261(_SX1261, SyncModem):
        pass

    class SX1262(_SX1262, SyncModem):
        pass

except ImportError:
    pass

try:
    from .async_modem import AsyncModem

    class AsyncSX1261(_SX1261, AsyncModem):
        pass

    class AsyncSX1262(_SX1262, AsyncModem):
        pass

except ImportError:
    pass



# Very basic test of ssl.SSLContext class.

try:
    import socket, ssl
except ImportError:
    print("SKIP")
    raise SystemExit

# Test constructing with arguments.
ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

# Test printing object.
ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
print("SSLContext" in str(ctx))

# Coverage test for destructor, and calling it twice.
if hasattr(ctx, "__del__"):
    ctx.__del__()
    ctx.__del__()

# Test calling .wrap_socket() method, multiple times.
ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
ctx.wrap_socket(socket.socket(), do_handshake_on_connect=False)
ctx.wrap_socket(socket.socket(), do_handshake_on_connect=False)



"""
This pre-processor parses a single file containing a list of
MP_REGISTER_ROOT_POINTER(variable declaration) items.

These are used to generate a header with the required entries for
"struct _mp_state_vm_t" in py/mpstate.h
"""

from __future__ import print_function

import argparse
import io
import re


PATTERN = re.compile(r"MP_REGISTER_ROOT_POINTER\((.*?)\);")


def find_root_pointer_registrations(filename):
    """Find any MP_REGISTER_ROOT_POINTER definitions in the provided file.

    :param str filename: path to file to check
    :return: List[variable_declaration]
    """
    with io.open(filename, encoding="utf-8") as c_file_obj:
        return set(re.findall(PATTERN, c_file_obj.read()))


def generate_root_pointer_header(root_pointers):
    """Generate header with root pointer entries.

    :param List[variable_declaration] root_pointers: root pointer declarations
    :return: None
    """

    # Print header file for all external modules.
    print("// Automatically generated by make_root_pointers.py.")
    print()

    for item in root_pointers:
        print(item, end=";")
        print()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("file", nargs=1, help="file with MP_REGISTER_ROOT_POINTER definitions")
    args = parser.parse_args()

    root_pointers = find_root_pointer_registrations(args.file[0])
    generate_root_pointer_header(sorted(root_pointers))


if __name__ == "__main__":
    main()



import sys
import unittest
from contextlib import closing, suppress, ExitStack


class ClosingTestCase(unittest.TestCase):
    class Closable:
        def __init__(self):
            self.closed = False

        def close(self):
            self.closed = True

    def test_closing(self):
        closable = self.Closable()
        with closing(closable) as c:
            self.assertFalse(c.closed)
        self.assertTrue(closable.closed)

    def test_closing_after_error(self):
        closable = self.Closable()
        exc = Exception()
        try:
            with closing(closable) as c:
                raise exc
        except Exception as e:
            self.assertEqual(exc, e)
        self.assertTrue(closable.closed)


class SuppressTestCase(unittest.TestCase):
    def test_suppress(self):
        with suppress(ValueError, TypeError):
            raise ValueError()
            raise TypeError()
        self.assertTrue(True)


class TestExitStack(unittest.TestCase):
    # @support.requires_docstrings
    def _test_instance_docs(self):
        # Issue 19330: ensure context manager instances have good docstrings
        cm_docstring = ExitStack.__doc__
        obj = ExitStack()
        self.assertEqual(obj.__doc__, cm_docstring)

    def test_no_resources(self):
        with ExitStack():
            pass

    def test_callback(self):
        expected = [
            ((), {}),
            ((1,), {}),
            ((1, 2), {}),
            ((), dict(example=1)),
            ((1,), dict(example=1)),
            ((1, 2), dict(example=1)),
        ]
        result = []

        def _exit(*args, **kwds):
            """Test metadata propagation"""
            result.append((args, kwds))

        with ExitStack() as stack:
            for args, kwds in reversed(expected):
                if args and kwds:
                    f = stack.callback(_exit, *args, **kwds)
                elif args:
                    f = stack.callback(_exit, *args)
                elif kwds:
                    f = stack.callback(_exit, **kwds)
                else:
                    f = stack.callback(_exit)
                self.assertIs(f, _exit)
            # for wrapper in stack._exit_callbacks:
            #     self.assertIs(wrapper.__wrapped__, _exit)
            #     self.assertNotEqual(wrapper.__name__, _exit.__name__)
            #     self.assertIsNone(wrapper.__doc__, _exit.__doc__)
        self.assertEqual(result, expected)

    def test_push(self):
        exc_raised = ZeroDivisionError

        def _expect_exc(exc_type, exc, exc_tb):
            self.assertIs(exc_type, exc_raised)

        def _suppress_exc(*exc_details):
            return True

        def _expect_ok(exc_type, exc, exc_tb):
            self.assertIsNone(exc_type)
            self.assertIsNone(exc)
            self.assertIsNone(exc_tb)

        class ExitCM(object):
            def __init__(self, check_exc):
                self.check_exc = check_exc

            def __enter__(self):
                self.fail("Should not be called!")

            def __exit__(self, *exc_details):
                self.check_exc(*exc_details)

        with ExitStack() as stack:
            stack.push(_expect_ok)
            self.assertIs(tuple(stack._exit_callbacks)[-1], _expect_ok)
            cm = ExitCM(_expect_ok)
            stack.push(cm)
            # self.assertIs(stack._exit_callbacks[-1].__self__, cm)
            stack.push(_suppress_exc)
            self.assertIs(tuple(stack._exit_callbacks)[-1], _suppress_exc)
            cm = ExitCM(_expect_exc)
            stack.push(cm)
            # self.assertIs(stack._exit_callbacks[-1].__self__, cm)
            stack.push(_expect_exc)
            self.assertIs(tuple(stack._exit_callbacks)[-1], _expect_exc)
            stack.push(_expect_exc)
            self.assertIs(tuple(stack._exit_callbacks)[-1], _expect_exc)
            1 / 0

    def test_enter_context(self):
        class TestCM(object):
            def __enter__(self):
                result.append(1)

            def __exit__(self, *exc_details):
                result.append(3)

        result = []
        cm = TestCM()
        with ExitStack() as stack:

            @stack.callback  # Registered first => cleaned up last
            def _exit():
                result.append(4)

            self.assertIsNotNone(_exit)
            stack.enter_context(cm)
            # self.assertIs(stack._exit_callbacks[-1].__self__, cm)
            result.append(2)
        self.assertEqual(result, [1, 2, 3, 4])

    def test_close(self):
        result = []
        with ExitStack() as stack:

            @stack.callback
            def _exit():
                result.append(1)

            self.assertIsNotNone(_exit)
            stack.close()
            result.append(2)
        self.assertEqual(result, [1, 2])

    def test_pop_all(self):
        result = []
        with ExitStack() as stack:

            @stack.callback
            def _exit():
                result.append(3)

            self.assertIsNotNone(_exit)
            new_stack = stack.pop_all()
            result.append(1)
        result.append(2)
        new_stack.close()
        self.assertEqual(result, [1, 2, 3])

    def test_exit_raise(self):
        with self.assertRaises(ZeroDivisionError):
            with ExitStack() as stack:
                stack.push(lambda *exc: False)
                1 / 0

    def test_exit_suppress(self):
        with ExitStack() as stack:
            stack.push(lambda *exc: True)
            1 / 0

    def test_exit_exception_chaining_reference(self):
        # Sanity check to make sure that ExitStack chaining matches
        # actual nested with statements
        exc_chain = []

        class RaiseExc:
            def __init__(self, exc):
                self.exc = exc

            def __enter__(self):
                return self

            def __exit__(self, *exc_details):
                exc_chain.append(exc_details[0])
                raise self.exc

        class RaiseExcWithContext:
            def __init__(self, outer, inner):
                self.outer = outer
                self.inner = inner

            def __enter__(self):
                return self

            def __exit__(self, *exc_details):
                try:
                    exc_chain.append(exc_details[0])
                    raise self.inner
                except:
                    exc_chain.append(sys.exc_info()[0])
                    raise self.outer

        class SuppressExc:
            def __enter__(self):
                return self

            def __exit__(self, *exc_details):
                type(self).saved_details = exc_details
                return True

        try:
            with RaiseExc(IndexError):
                with RaiseExcWithContext(KeyError, AttributeError):
                    with SuppressExc():
                        with RaiseExc(ValueError):
                            1 / 0
        except IndexError as exc:
            # self.assertIsInstance(exc.__context__, KeyError)
            # self.assertIsInstance(exc.__context__.__context__, AttributeError)
            # Inner exceptions were suppressed
            # self.assertIsNone(exc.__context__.__context__.__context__)
            exc_chain.append(type(exc))
            assert tuple(exc_chain) == (
                ZeroDivisionError,
                None,
                AttributeError,
                KeyError,
                IndexError,
            )
        else:
            self.fail("Expected IndexError, but no exception was raised")
        # Check the inner exceptions
        inner_exc = SuppressExc.saved_details[1]
        self.assertIsInstance(inner_exc, ValueError)
        # self.assertIsInstance(inner_exc.__context__, ZeroDivisionError)

    def test_exit_exception_chaining(self):
        # Ensure exception chaining matches the reference behaviour
        exc_chain = []

        def raise_exc(exc):
            frame_exc = sys.exc_info()[0]
            if frame_exc is not None:
                exc_chain.append(frame_exc)
            exc_chain.append(exc)
            raise exc

        saved_details = None

        def suppress_exc(*exc_details):
            nonlocal saved_details
            saved_details = exc_details
            assert exc_chain[-1] == exc_details[0]
            exc_chain[-1] = None
            return True

        try:
            with ExitStack() as stack:
                stack.callback(raise_exc, IndexError)
                stack.callback(raise_exc, KeyError)
                stack.callback(raise_exc, AttributeError)
                stack.push(suppress_exc)
                stack.callback(raise_exc, ValueError)
                1 / 0
        except IndexError as exc:
            # self.assertIsInstance(exc.__context__, KeyError)
            # self.assertIsInstance(exc.__context__.__context__, AttributeError)
            # Inner exceptions were suppressed
            # self.assertIsNone(exc.__context__.__context__.__context__)
            assert tuple(exc_chain) == (
                ZeroDivisionError,
                None,
                AttributeError,
                KeyError,
                IndexError,
            )
        else:
            self.fail("Expected IndexError, but no exception was raised")
        # Check the inner exceptions
        inner_exc = saved_details[1]
        self.assertIsInstance(inner_exc, ValueError)
        # self.assertIsInstance(inner_exc.__context__, ZeroDivisionError)

    def test_exit_exception_non_suppressing(self):
        # http://bugs.python.org/issue19092
        def raise_exc(exc):
            raise exc

        def suppress_exc(*exc_details):
            return True

        try:
            with ExitStack() as stack:
                stack.callback(lambda: None)
                stack.callback(raise_exc, IndexError)
        except Exception as exc:
            self.assertIsInstance(exc, IndexError)
        else:
            self.fail("Expected IndexError, but no exception was raised")

        try:
            with ExitStack() as stack:
                stack.callback(raise_exc, KeyError)
                stack.push(suppress_exc)
                stack.callback(raise_exc, IndexError)
        except Exception as exc:
            self.assertIsInstance(exc, KeyError)
        else:
            self.fail("Expected KeyError, but no exception was raised")

    def _test_exit_exception_with_correct_context(self):
        # http://bugs.python.org/issue20317
        @contextmanager
        def gets_the_context_right(exc):
            try:
                yield
            finally:
                raise exc

        exc1 = Exception(1)
        exc2 = Exception(2)
        exc3 = Exception(3)
        exc4 = Exception(4)

        # The contextmanager already fixes the context, so prior to the
        # fix, ExitStack would try to fix it *again* and get into an
        # infinite self-referential loop
        try:
            with ExitStack() as stack:
                stack.enter_context(gets_the_context_right(exc4))
                stack.enter_context(gets_the_context_right(exc3))
                stack.enter_context(gets_the_context_right(exc2))
                raise exc1
        except Exception as exc:
            self.assertIs(exc, exc4)
            self.assertIs(exc.__context__, exc3)
            self.assertIs(exc.__context__.__context__, exc2)
            self.assertIs(exc.__context__.__context__.__context__, exc1)
            self.assertIsNone(exc.__context__.__context__.__context__.__context__)

    def _test_exit_exception_with_existing_context(self):
        # Addresses a lack of test coverage discovered after checking in a
        # fix for issue 20317 that still contained debugging code.
        def raise_nested(inner_exc, outer_exc):
            try:
                raise inner_exc
            finally:
                raise outer_exc

        exc1 = Exception(1)
        exc2 = Exception(2)
        exc3 = Exception(3)
        exc4 = Exception(4)
        exc5 = Exception(5)
        try:
            with ExitStack() as stack:
                stack.callback(raise_nested, exc4, exc5)
                stack.callback(raise_nested, exc2, exc3)
                raise exc1
        except Exception as exc:
            self.assertIs(exc, exc5)
            self.assertIs(exc.__context__, exc4)
            self.assertIs(exc.__context__.__context__, exc3)
            self.assertIs(exc.__context__.__context__.__context__, exc2)
            self.assertIs(exc.__context__.__context__.__context__.__context__, exc1)
            self.assertIsNone(exc.__context__.__context__.__context__.__context__.__context__)

    def test_body_exception_suppress(self):
        def suppress_exc(*exc_details):
            return True

        try:
            with ExitStack() as stack:
                stack.push(suppress_exc)
                1 / 0
        except IndexError as exc:
            self.fail("Expected no exception, got IndexError")

    def test_exit_exception_chaining_suppress(self):
        with ExitStack() as stack:
            stack.push(lambda *exc: True)
            stack.push(lambda *exc: 1 / 0)
            stack.push(lambda *exc: {}[1])

    def test_excessive_nesting(self):
        # The original implementation would die with RecursionError here
        with ExitStack() as stack:
            for i in range(5000):
                stack.callback(int)

    def test_instance_bypass(self):
        class Example(object):
            pass

        cm = Example()
        cm.__exit__ = object()
        stack = ExitStack()
        self.assertRaises(AttributeError, stack.enter_context, cm)
        stack.push(cm)
        self.assertIs(tuple(stack._exit_callbacks)[-1], cm)


if __name__ == "__main__":
    unittest.main()



# Test Event class

try:
    import asyncio
except ImportError:
    print("SKIP")
    raise SystemExit


async def task(id, ev):
    print("start", id)
    print(await ev.wait())
    print("end", id)


async def task_delay_set(t, ev):
    await asyncio.sleep(t)
    print("set event")
    ev.set()


async def main():
    ev = asyncio.Event()

    # Set and clear without anything waiting, and test is_set()
    print(ev.is_set())
    ev.set()
    print(ev.is_set())
    ev.clear()
    print(ev.is_set())

    # Create 2 tasks waiting on the event
    print("----")
    asyncio.create_task(task(1, ev))
    asyncio.create_task(task(2, ev))
    print("yield")
    await asyncio.sleep(0)
    print("set event")
    ev.set()
    print("yield")
    await asyncio.sleep(0)

    # Create a task waiting on the already-set event
    print("----")
    asyncio.create_task(task(3, ev))
    print("yield")
    await asyncio.sleep(0)

    # Clear event, start a task, then set event again
    print("----")
    print("clear event")
    ev.clear()
    asyncio.create_task(task(4, ev))
    await asyncio.sleep(0)
    print("set event")
    ev.set()
    await asyncio.sleep(0)

    # Cancel a task waiting on an event (set event then cancel task)
    print("----")
    ev = asyncio.Event()
    t = asyncio.create_task(task(5, ev))
    await asyncio.sleep(0)
    ev.set()
    t.cancel()
    await asyncio.sleep(0.1)

    # Cancel a task waiting on an event (cancel task then set event)
    print("----")
    ev = asyncio.Event()
    t = asyncio.create_task(task(6, ev))
    await asyncio.sleep(0)
    t.cancel()
    ev.set()
    await asyncio.sleep(0.1)

    # Wait for an event that does get set in time
    print("----")
    ev.clear()
    asyncio.create_task(task_delay_set(0.01, ev))
    await asyncio.wait_for(ev.wait(), 0.1)
    await asyncio.sleep(0)

    # Wait for an event that doesn't get set in time
    print("----")
    ev.clear()
    asyncio.create_task(task_delay_set(0.1, ev))
    try:
        await asyncio.wait_for(ev.wait(), 0.01)
    except asyncio.TimeoutError:
        print("TimeoutError")
    await ev.wait()


asyncio.run(main())



# Test performance of importing an .mpy file many times.

import sys, io, vfs

if not hasattr(io, "IOBase"):
    print("SKIP")
    raise SystemExit

# This is the test.py file that is compiled to test.mpy below.
"""
class A:
    def __init__(self, arg):
        self.arg = arg
    def write(self):
        pass
    def read(self):
        pass
def f():
    print, str, bytes, dict
    Exception, ValueError, TypeError
    x = "this will be a string object"
    x = b"this will be a bytes object"
    x = ("const tuple", None, False, True, 1, 2, 3)
result = 123
"""
file_data = b'M\x06\x00\x1f\x14\x03\x0etest.py\x00\x0f\x02A\x00\x02f\x00\x0cresult\x00/-5#\x82I\x81{\x81w\x82/\x81\x05\x81\x17Iom\x82\x13\x06arg\x00\x05\x1cthis will be a string object\x00\x06\x1bthis will be a bytes object\x00\n\x07\x05\x0bconst tuple\x00\x01\x02\x03\x07\x011\x07\x012\x07\x013\x81\\\x10\n\x01\x89\x07d`T2\x00\x10\x024\x02\x16\x022\x01\x16\x03"\x80{\x16\x04Qc\x02\x81d\x00\x08\x02(DD\x11\x05\x16\x06\x10\x02\x16\x072\x00\x16\x082\x01\x16\t2\x02\x16\nQc\x03`\x1a\x08\x08\x12\x13@\xb1\xb0\x18\x13Qc@\t\x08\t\x12` Qc@\t\x08\n\x12``Qc\x82@ \x0e\x03\x80\x08+)##\x12\x0b\x12\x0c\x12\r\x12\x0e*\x04Y\x12\x0f\x12\x10\x12\x11*\x03Y#\x00\xc0#\x01\xc0#\x02\xc0Qc'


class File(io.IOBase):
    def __init__(self):
        self.off = 0

    def ioctl(self, request, arg):
        return 0

    def readinto(self, buf):
        buf[:] = memoryview(file_data)[self.off : self.off + len(buf)]
        self.off += len(buf)
        return len(buf)


class FS:
    def mount(self, readonly, mkfs):
        pass

    def chdir(self, path):
        pass

    def stat(self, path):
        if path == "/__injected.mpy":
            return tuple(0 for _ in range(10))
        else:
            raise OSError(-2)  # ENOENT

    def open(self, path, mode):
        return File()


def mount():
    vfs.mount(FS(), "/__remote")
    sys.path.insert(0, "/__remote")


def test(r):
    global result
    for _ in r:
        sys.modules.clear()
        module = __import__("__injected")
    result = module.result


###########################################################################
# Benchmark interface

bm_params = {
    (32, 10): (50,),
    (1000, 10): (500,),
    (5000, 10): (5000,),
}


def bm_setup(params):
    (nloop,) = params
    mount()
    return lambda: test(range(nloop)), lambda: (nloop, result)



# Example using PIO to wait for a pin change and raise an IRQ.
#
# Demonstrates:
#   - PIO wrapping
#   - PIO wait instruction, waiting on an input pin
#   - PIO irq instruction, in blocking mode with relative IRQ number
#   - setting the in_base pin for a StateMachine
#   - setting an irq handler for a StateMachine
#   - instantiating 2x StateMachine's with the same program and different pins

# ruff: noqa: F821 - @asm_pio decorator adds names to function scope

import time
from machine import Pin
import rp2


@rp2.asm_pio()
def wait_pin_low():
    wrap_target()

    wait(0, pin, 0)
    irq(block, rel(0))
    wait(1, pin, 0)

    wrap()


def handler(sm):
    # Print a (wrapping) timestamp, and the state machine object.
    print(time.ticks_ms(), sm)


# Instantiate StateMachine(0) with wait_pin_low program on Pin(16).
pin16 = Pin(16, Pin.IN, Pin.PULL_UP)
sm0 = rp2.StateMachine(0, wait_pin_low, in_base=pin16)
sm0.irq(handler)

# Instantiate StateMachine(1) with wait_pin_low program on Pin(17).
pin17 = Pin(17, Pin.IN, Pin.PULL_UP)
sm1 = rp2.StateMachine(1, wait_pin_low, in_base=pin17)
sm1.irq(handler)

# Start the StateMachine's running.
sm0.active(1)
sm1.active(1)

# Now, when Pin(16) or Pin(17) is pulled low a message will be printed to the REPL.



# delete local then try to reference it
def f():
    x = 1
    y = 2
    print(x, y)
    del x
    print(y)
    try:
        print(x)
    except NameError:
        print("NameError");
f()

# delete local then try to delete it again
def g():
    x = 3
    y = 4
    print(x, y)
    del x
    print(y)
    try:
        del x
    except NameError:
        print("NameError");
g()



try:
    import machine

    machine.PinBase
except:
    print("SKIP")
    raise SystemExit


class MyPin(machine.PinBase):
    def __init__(self):
        print("__init__")
        self.v = False

    def value(self, v=None):
        print("value:", v)
        if v is None:
            self.v = not self.v
            return int(self.v)


p = MyPin()

print(p.value())
print(p.value())
print(p.value())
p.value(1)
p.value(0)



"""
UART test for the CC3200 based boards.
UART0 and UART1 must be connected together for this test to pass.
"""

from machine import UART
from machine import Pin
import os
import time

mch = os.uname().machine
if "LaunchPad" in mch:
    uart_id_range = range(0, 2)
    uart_pins = [
        [("GP12", "GP13"), ("GP12", "GP13", "GP7", "GP6")],
        [("GP16", "GP17"), ("GP16", "GP17", "GP7", "GP6")],
    ]
elif "WiPy" in mch:
    uart_id_range = range(0, 2)
    uart_pins = [
        [("GP12", "GP13"), ("GP12", "GP13", "GP7", "GP6")],
        [("GP16", "GP17"), ("GP16", "GP17", "GP7", "GP6")],
    ]
else:
    raise Exception("Board not supported!")

# just in case we have the repl duplicated on any of the uarts
os.dupterm(None)

for uart_id in uart_id_range:
    uart = UART(uart_id, 38400)
    print(uart)
    uart.init(57600, 8, None, 1, pins=uart_pins[uart_id][0])
    uart.init(baudrate=9600, stop=2, parity=UART.EVEN, pins=uart_pins[uart_id][1])
    uart.init(baudrate=115200, parity=UART.ODD, stop=0, pins=uart_pins[uart_id][0])
    uart = UART(baudrate=1000000)
    uart.sendbreak()

uart = UART(baudrate=1000000)
uart = UART()
print(uart)
uart = UART(baudrate=38400, pins=("GP12", "GP13"))
print(uart)
uart = UART(pins=("GP12", "GP13"))
print(uart)
uart = UART(pins=(None, "GP17"))
print(uart)
uart = UART(baudrate=57600, pins=("GP16", "GP17"))
print(uart)

# now it's time for some loopback tests between the uarts
uart0 = UART(0, 1000000, pins=uart_pins[0][0])
print(uart0)
uart1 = UART(1, 1000000, pins=uart_pins[1][0])
print(uart1)

print(uart0.write(b"123456") == 6)
print(uart1.read() == b"123456")

print(uart1.write(b"123") == 3)
print(uart0.read(1) == b"1")
print(uart0.read(2) == b"23")
print(uart0.read() == None)

uart0.write(b"123")
buf = bytearray(3)
print(uart1.readinto(buf, 1) == 1)
print(buf)
print(uart1.readinto(buf) == 2)
print(buf)

# try initializing without the id
uart0 = UART(baudrate=1000000, pins=uart_pins[0][0])
uart0.write(b"1234567890")
time.sleep_ms(2)  # because of the fifo interrupt levels
print(uart1.any() == 10)
print(uart1.readline() == b"1234567890")
print(uart1.any() == 0)

uart0.write(b"1234567890")
print(uart1.read() == b"1234567890")

# tx only mode
uart0 = UART(0, 1000000, pins=("GP12", None))
print(uart0.write(b"123456") == 6)
print(uart1.read() == b"123456")
print(uart1.write(b"123") == 3)
print(uart0.read() == None)

# rx only mode
uart0 = UART(0, 1000000, pins=(None, "GP13"))
print(uart0.write(b"123456") == 6)
print(uart1.read() == None)
print(uart1.write(b"123") == 3)
print(uart0.read() == b"123")

# leave pins as they were (rx only mode)
uart0 = UART(0, 1000000, pins=None)
print(uart0.write(b"123456") == 6)
print(uart1.read() == None)
print(uart1.write(b"123") == 3)
print(uart0.read() == b"123")

# no pin assignment
uart0 = UART(0, 1000000, pins=(None, None))
print(uart0.write(b"123456789") == 9)
print(uart1.read() == None)
print(uart1.write(b"123456789") == 9)
print(uart0.read() == None)
print(Pin.board.GP12)
print(Pin.board.GP13)

# check for memory leaks...
for i in range(0, 1000):
    uart0 = UART(0, 1000000)
    uart1 = UART(1, 1000000)

# next ones must raise
try:
    UART(0, 9600, parity=None, pins=("GP12", "GP13", "GP7"))
except Exception:
    print("Exception")

try:
    UART(0, 9600, parity=UART.ODD, pins=("GP12", "GP7"))
except Exception:
    print("Exception")

uart0 = UART(0, 1000000)
uart0.deinit()
try:
    uart0.any()
except Exception:
    print("Exception")

try:
    uart0.read()
except Exception:
    print("Exception")

try:
    uart0.write("abc")
except Exception:
    print("Exception")

try:
    uart0.sendbreak("abc")
except Exception:
    print("Exception")

try:
    UART(2, 9600)
except Exception:
    print("Exception")

for uart_id in uart_id_range:
    uart = UART(uart_id, 1000000)
    uart.deinit()
    # test printing an uninitialized uart
    print(uart)
    # initialize it back and check that it works again
    uart.init(115200)
    print(uart)
    uart.read()



# test the special functions imported from math

try:
    from math import *

    erf
except (ImportError, NameError):
    print("SKIP")
    raise SystemExit

test_values = [
    -8.0,
    -2.5,
    -1,
    -0.5,
    0.0,
    0.5,
    2.5,
    8.0,
]
pos_test_values = [
    0.001,
    0.1,
    0.5,
    1.0,
    1.5,
    10.0,
]

functions = [
    ("expm1", expm1, test_values),
    ("log2", log2, test_values),
    ("log10", log10, test_values),
    ("cosh", cosh, test_values),
    ("sinh", sinh, test_values),
    ("tanh", tanh, [-1e6, -100] + test_values + [100, 1e6]),
    ("acosh", acosh, [1.0, 5.0, 1.0]),
    ("asinh", asinh, test_values),
    ("atanh", atanh, [-0.99, -0.5, 0.0, 0.5, 0.99]),
    ("erf", erf, test_values),
    ("erfc", erfc, test_values),
    ("gamma", gamma, pos_test_values),
    ("lgamma", lgamma, pos_test_values + [50.0, 100.0]),
]

for function_name, function, test_vals in functions:
    for value in test_vals:
        try:
            ans = "{:.4g}".format(function(value))
        except ValueError as e:
            ans = str(e)
        print("{}({:.4g}) = {}".format(function_name, value, ans))



from machine import Pin, Signal

# 96Boards/Qualcomm DragonBoard 410c
#
# By default, on-board LEDs are controlled by kernel LED driver.
# To make corresponding pins be available as normal GPIO,
# corresponding driver needs to be unbound first (as root):
# echo -n "soc:leds" >/sys/class/leds/apq8016-sbc:green:user1/device/driver/unbind
# Note that application also either should be run as root, or
# /sys/class/gpio ownership needs to be changed.
# Likewise, onboard buttons are controlled by gpio_keys driver.
# To release corresponding GPIOs:
# echo -n "gpio_keys" >/sys/class/input/input1/device/driver/unbind

# User LED 1 on gpio21
LED = Signal(Pin(21, Pin.OUT))

# User LED 2 on gpio120
LED2 = Signal(Pin(120, Pin.OUT))

# Button S3 on gpio107
BUTTON = Pin(107, Pin.IN)
